"""Indices module."""

from __future__ import annotations

from xclim.indices._agro import *
from xclim.indices._anuclim import *
from xclim.indices._conversion import *
from xclim.indices._hydrology import *
from xclim.indices._multivariate import *
from xclim.indices._simple import *
from xclim.indices._synoptic import *
from xclim.indices._threshold import *
from xclim.indices.fire import (
    cffwis_indices,
    drought_code,
    duff_moisture_code,
    fire_season,
    griffiths_drought_factor,
    keetch_byram_drought_index,
    mcarthur_forest_fire_danger_index,
)

# pylint: disable=pointless-string-statement
"""
Notes for docstrings
--------------------

The docstrings adhere to the `NumPy`_ style convention and is meant as a way to store CF-Convention metadata as
well as information relevant to third party libraries (such as a WPS server).

The first line of the docstring (the short summary), will be assigned to the output's `long_name` attribute. The
`long_name` attribute is defined by the NetCDF User Guide to contain a long descriptive name which may, for example,
be used for labeling plots

The second paragraph will be considered as the "*abstract*", or the CF global "*comment*" (miscellaneous information
about the data or methods used to produce it).

The third and fourth sections are the **Parameters** and **Returns** sections describing the input and output values
respectively.

The following example shows the structure of an indice definition:

.. code-block:: python

   @declare_units(var1="[units dimension]", thresh="[units dimension]")
   def indice_name(var1: xr.DataArray, thresh: str = "0 degC", freq: str = "YS"):
       \"\"\"
       The first line is the title

       The second paragraph is the abstract.

       Parameters
       ----------
       var1 : xarray.DataArray
         Description of variable (no need to specify units, the signature and decorator carry this information).
         <var1> is a short name like "tas", "pr" or "sfcWind".
       thresh : str
         Description of the threshold (no need to specify units or the default again).
         Parameters required to run the computation must always have a working default value.
       freq : str
         Resampling frequency. (the signature carries the default value information)

       Returns
       -------
       <var_name> : xarray.DataArray, [output units dimension]
         Output's <long_name>
         Specifying <var_name> is optional.
       \"\"\"
       <body of the function>
       # Don't forget to explicitly handle the units!

The next sections would be **Notes** and **References**:

.. code-block:: python

    Notes
    -----
    This is where the mathematical equation is described.
    At the end of the description, convention suggests
    to add a reference [example]_:

        .. math::

            3987^12 + 4365^12 = 4472^12

    References
    ----------
    .. [example] Smith, T.J. and Huard, D. (2018). "CF Docstrings:
        A manifesto on conventions and the metaphysical nature
        of ontological python documentation." Climate Aesthetics,
        vol. 1, pp. 121-155.

Indice Descriptions
===================
.. _`NumPy`: https://numpydoc.readthedocs.io/en/latest/format.html#docstring-standard
"""
# pylint: enable=pointless-string-statement

# TODO: Should we reference the standard vocabulary we're using ?
# E.g. http://vocab.nerc.ac.uk/collection/P07/current/BHMHISG2/



"""Atmospheric conversion definitions."""

from __future__ import annotations

from xclim import indices
from xclim.core.cfchecks import cfcheck_from_name
from xclim.core.indicator import Indicator
from xclim.core.utils import InputKind

__all__ = [
    "corn_heat_units",
    "heat_index",
    "humidex",
    "longwave_upwelling_radiation_from_net_downwelling",
    "mean_radiant_temperature",
    "potential_evapotranspiration",
    "rain_approximation",
    "relative_humidity",
    "relative_humidity_from_dewpoint",
    "saturation_vapor_pressure",
    "shortwave_upwelling_radiation_from_net_downwelling",
    "snowfall_approximation",
    "specific_humidity",
    "specific_humidity_from_dewpoint",
    "tg",
    "universal_thermal_climate_index",
    "water_budget",
    "water_budget_from_tas",
    "wind_chill_index",
    "wind_power_potential",
    "wind_profile",
    "wind_speed_from_vector",
    "wind_vector_from_speed",
]


class Converter(Indicator):
    """Class for indicators doing variable conversion (dimension-independent 1-to-1 computation)."""

    def cfcheck(self, **das):
        for varname, vardata in das.items():
            try:
                # Only check standard_name, and not cell_methods which depends on the variable's frequency.
                cfcheck_from_name(varname, vardata, attrs=["standard_name"])
            except KeyError:  # noqa: S110
                # Silently ignore unknown variables.
                pass


humidex = Converter(
    title="Humidex",
    identifier="humidex",
    units="C",
    standard_name="air_temperature",
    long_name="Humidex index",
    description="Humidex index describing the temperature felt by the average person in response to relative humidity.",
    cell_methods="",
    keywords="heatwave",
    abstract="The humidex describes the temperature felt by a person when relative humidity is taken into account. "
    "It can be interpreted as the equivalent temperature felt when the air is dry.",
    compute=indices.humidex,
)


heat_index = Converter(
    title="Heat index",
    identifier="heat_index",
    units="C",
    standard_name="air_temperature",
    long_name="Heat index",
    description="Perceived temperature after relative humidity is taken into account.",
    cell_methods="",
    abstract="The heat index is an estimate of the temperature felt by a person in the shade "
    "when relative humidity is taken into account.",
    compute=indices.heat_index,
)


tg = Converter(
    title="Mean temperature",
    identifier="tg",
    units="K",
    standard_name="air_temperature",
    long_name="Daily mean temperature",
    description="Estimated mean temperature from maximum and minimum temperatures.",
    cell_methods="time: mean within days",
    abstract="The average daily temperature assuming a symmetrical temperature distribution (Tg = (Tx + Tn) / 2).",
    compute=indices.tas,
)


wind_speed_from_vector = Converter(
    title="Wind speed and direction from vector",
    identifier="wind_speed_from_vector",
    var_name=["sfcWind", "sfcWindfromdir"],
    units=["m s-1", "degree"],
    standard_name=["wind_speed", "wind_from_direction"],
    description=[
        "Wind speed computed as the magnitude of the (uas, vas) vector.",
        "Wind direction computed as the angle of the (uas, vas) vector."
        " A direction of 0° is attributed to winds with a speed under {calm_wind_thresh}.",
    ],
    long_name=["Near-surface wind speed", "Near-surface wind from direction"],
    cell_methods="",
    abstract="Calculation of the magnitude and direction of the wind speed "
    "from the two components west-east and south-north.",
    compute=indices.uas_vas_2_sfcwind,
)


wind_vector_from_speed = Converter(
    title="Wind vector from speed and direction",
    identifier="wind_vector_from_speed",
    var_name=["uas", "vas"],
    units=["m s-1", "m s-1"],
    standard_name=["eastward_wind", "northward_wind"],
    long_name=["Near-surface eastward wind", "Near-surface northward wind"],
    description=[
        "Eastward wind speed computed from the magnitude of its speed and direction of origin.",
        "Northward wind speed computed from magnitude of its speed and direction of origin.",
    ],
    cell_methods="",
    abstract="Calculation of the two components (west-east and north-south) of the wind "
    "from the magnitude of its speed and direction of origin.",
    compute=indices.sfcwind_2_uas_vas,
)

wind_power_potential = Converter(
    title="Wind power potential",
    identifier="wind_power_potential",
    units="",
    long_name="Wind power potential",
    description="Wind power potential using a semi-idealized turbine power curve using a cut_in speed of {cut_in}, "
    "a rated speed of {rated}, and a cut_out speed of {cut_out}.",
    cell_methods="",
    abstract="Calculation of the wind power potential using a semi-idealized turbine power curve.",
    compute=indices.wind_power_potential,
)

wind_profile = Converter(
    title="Wind profile",
    identifier="wind_profile",
    var_name="wind_speed",
    units="m s-1",
    standard_name="wind_speed",
    long_name="Wind speed at height {h}",
    description="Wind speed at a height of {h} computed from the wind speed at {h_r} using a power law profile.",
    cell_methods="",
    abstract="Calculation of the wind speed at a given height from the wind speed at a reference height.",
    compute=indices.wind_profile,
)

saturation_vapor_pressure = Converter(
    title="Saturation vapour pressure (e_sat)",
    identifier="e_sat",
    units="Pa",
    long_name='Saturation vapour pressure ("{method}" method)',
    description=lambda **kws: (
        "The saturation vapour pressure was calculated from a temperature according to the {method} method."
    )
    + (
        " The computation was done in reference to ice for temperatures below {ice_thresh}."
        if kws["ice_thresh"] is not None
        else ""
    ),
    abstract="Calculation of the saturation vapour pressure from the temperature, according to a given method. "
    "If ice_thresh is given, the calculation is done with reference to ice for temperatures below this threshold.",
    compute=indices.saturation_vapor_pressure,
)


relative_humidity_from_dewpoint = Converter(
    title="Relative humidity from temperature and dewpoint temperature",
    identifier="hurs_fromdewpoint",
    units="%",
    var_name="hurs",
    long_name='Relative humidity ("{method}" method)',
    standard_name="relative_humidity",
    description=lambda **kws: (
        "Computed from temperature, and dew point temperature through the "
        "saturation vapour pressures, which were calculated "
        "according to the {method} method."
    )
    + (
        " The computation was done in reference to ice for temperatures below {ice_thresh}."
        if kws["ice_thresh"] is not None
        else ""
    ),
    abstract="Calculation of relative humidity from temperature and dew point using the saturation vapour pressure.",
    compute=indices.relative_humidity,
    parameters={
        "tdps": {"kind": InputKind.VARIABLE},
        "huss": None,
        "ps": None,
        "invalid_values": "mask",
    },
)


relative_humidity = Converter(
    title="Relative humidity from temperature, specific humidity, and pressure",
    identifier="hurs",
    units="%",
    var_name="hurs",
    long_name='Relative Humidity ("{method}" method)',
    standard_name="relative_humidity",
    description=lambda **kws: (
        "Computed from temperature, specific humidity and pressure through the saturation vapour pressure, "
        "which was calculated from temperature according to the {method} method."
    )
    + (
        " The computation was done in reference to ice for temperatures below {ice_thresh}."
        if kws["ice_thresh"] is not None
        else ""
    ),
    abstract="Calculation of relative humidity from temperature, "
    "specific humidity, and pressure using the saturation vapour pressure.",
    compute=indices.relative_humidity,
    parameters={
        "tdps": None,
        "huss": {"kind": InputKind.VARIABLE},
        "ps": {"kind": InputKind.VARIABLE},
        "invalid_values": "mask",
    },
)


specific_humidity = Converter(
    title="Specific humidity from temperature, relative humidity, and pressure",
    identifier="huss",
    units="",
    var_name="huss",
    long_name='Specific Humidity ("{method}" method)',
    standard_name="specific_humidity",
    description=lambda **kws: (
        "Computed from temperature, relative humidity and pressure through the saturation vapour pressure, "
        "which was calculated from temperature according to the {method} method."
    )
    + (
        " The computation was done in reference to ice for temperatures below {ice_thresh}."
        if kws["ice_thresh"] is not None
        else ""
    ),
    abstract="Calculation of specific humidity from temperature, "
    "relative humidity, and pressure using the saturation vapour pressure.",
    compute=indices.specific_humidity,
    parameters={"invalid_values": "mask"},
)

specific_humidity_from_dewpoint = Converter(
    title="Specific humidity from dew point temperature and pressure",
    identifier="huss_fromdewpoint",
    units="",
    long_name='Specific humidity ("{method}" method)',
    standard_name="specific_humidity",
    description=(
        "Computed from dewpoint temperature and pressure through the saturation "
        "vapor pressure, which was calculated according to the {method} method."
    ),
    abstract="Calculation of the specific humidity from dew point temperature "
    "and pressure using the saturation vapour pressure.",
    compute=indices.specific_humidity_from_dewpoint,
)

snowfall_approximation = Converter(
    title="Snowfall approximation",
    identifier="prsn",
    units="kg m-2 s-1",
    standard_name="solid_precipitation_flux",
    long_name='Solid precipitation ("{method}" method with temperature at or below {thresh})',
    description=(
        "Solid precipitation estimated from total precipitation and temperature"
        " with method {method} and threshold temperature {thresh}."
    ),
    abstract="Solid precipitation estimated from total precipitation and temperature "
    "with a given method and temperature threshold.",
    compute=indices.snowfall_approximation,
    context="hydro",
)


rain_approximation = Converter(
    title="Rainfall approximation",
    identifier="prlp",
    units="kg m-2 s-1",
    standard_name="precipitation_flux",
    long_name='Liquid precipitation ("{method}" method with temperature at or above {thresh})',
    description=(
        "Liquid precipitation estimated from total precipitation and temperature"
        " with method {method} and threshold temperature {thresh}."
    ),
    abstract="Liquid precipitation estimated from total precipitation and temperature "
    "with a given method and temperature threshold.",
    compute=indices.rain_approximation,
    context="hydro",
)


wind_chill_index = Converter(
    title="Wind chill",
    identifier="wind_chill",
    units="degC",
    long_name="Wind chill factor",
    description=lambda **kws: (
        "Wind chill index describing the temperature felt by the average person in response to cold wind."
    )
    + (
        "A slow-wind version of the wind chill index was used for wind speeds under 5 km/h and invalid "
        "temperatures were masked (T > 0°C)."
        if kws["method"] == "CAN"
        else "Invalid temperatures (T > 50°F) and winds (V < 3 mph) where masked."
    ),
    abstract="Wind chill factor is an index that equates to how cold an average person feels. "
    "It is calculated from the temperature and the wind speed at 10 m. "
    "As defined by Environment and Climate Change Canada, a second formula is used for light winds. "
    "The standard formula is otherwise the same as used in the United States.",
    compute=indices.wind_chill_index,
    parameters={"mask_invalid": True},
)


potential_evapotranspiration = Converter(
    title="Potential evapotranspiration",
    identifier="potential_evapotranspiration",
    var_name="evspsblpot",
    units="kg m-2 s-1",
    standard_name="water_potential_evapotranspiration_flux",
    long_name='Potential evapotranspiration ("{method}" method)',
    description=(
        "The potential for water evaporation from soil and transpiration by plants if the water "
        "supply is sufficient, calculated with the {method} method."
    ),
    abstract=(
        "The potential for water evaporation from soil and transpiration by plants if the water "
        "supply is sufficient, calculated with a given method."
    ),
    compute=indices.potential_evapotranspiration,
)

water_budget_from_tas = Converter(
    title="Water budget",
    identifier="water_budget_from_tas",
    units="kg m-2 s-1",
    long_name='Water budget ("{method}" method)',
    description=(
        "Precipitation minus potential evapotranspiration as a measure of an approximated surface water budget, "
        "where the potential evapotranspiration is calculated with the {method} method."
    ),
    abstract=(
        "Precipitation minus potential evapotranspiration as a measure of an approximated surface water budget, "
        "where the potential evapotranspiration is calculated with a given method."
    ),
    compute=indices.water_budget,
)

water_budget = Converter(
    title="Water budget",
    identifier="water_budget",
    units="kg m-2 s-1",
    long_name="Water budget",
    description=(
        "Precipitation minus potential evapotranspiration as a measure of an approximated surface water budget."
    ),
    abstract=(
        "Precipitation minus potential evapotranspiration as a measure of an approximated surface water budget."
    ),
    compute=indices.water_budget,
    parameters={"method": "dummy"},
)


corn_heat_units = Converter(
    title="Corn heat units",
    identifier="corn_heat_units",
    units="",
    long_name="Corn heat units (Tmin > {thresh_tasmin} and Tmax > {thresh_tasmax})",
    description="Temperature-based index used to estimate the development of corn crops. "
    "Corn growth occurs when the minimum and maximum daily temperatures both exceed "
    "{thresh_tasmin} and {thresh_tasmax}, respectively.",
    abstract="A temperature-based index used to estimate the development of corn crops. "
    "Corn growth occurs when the daily minimum and maximum temperatures exceed given thresholds.",
    var_name="chu",
    cell_methods="",
    missing="skip",
    compute=indices.corn_heat_units,
)

universal_thermal_climate_index = Converter(
    title="Universal Thermal Climate Index (UTCI)",
    identifier="utci",
    units="K",
    long_name="Universal Thermal Climate Index (UTCI)",
    description="UTCI is the equivalent temperature for the environment derived from a reference environment "
    "and is used to evaluate heat stress in outdoor spaces.",
    abstract="UTCI is the equivalent temperature for the environment derived from a reference environment "
    "and is used to evaluate heat stress in outdoor spaces.",
    cell_methods="",
    var_name="utci",
    compute=indices.universal_thermal_climate_index,
)

mean_radiant_temperature = Converter(
    title="Mean radiant temperature",
    identifier="mean_radiant_temperature",
    units="K",
    long_name="Mean radiant temperature",
    description="The incidence of radiation on the body from all directions.",
    abstract="The average temperature of solar and thermal radiation incident on the body's exterior.",
    cell_methods="",
    var_name="mrt",
    compute=indices.mean_radiant_temperature,
)


shortwave_upwelling_radiation_from_net_downwelling = Converter(
    title="Upwelling shortwave radiation",
    identifier="shortwave_upwelling_radiation_from_net_downwelling",
    units="W m-2",
    standard_name="surface_upwelling_shortwave_flux",
    long_name="Upwelling shortwave flux",
    description="The calculation of upwelling shortwave radiative flux from net surface shortwave "
    "and downwelling surface shortwave fluxes.",
    var_name="rsus",
    compute=indices.shortwave_upwelling_radiation_from_net_downwelling,
)

longwave_upwelling_radiation_from_net_downwelling = Converter(
    title="Upwelling longwave radiation",
    identifier="longwave_upwelling_radiation_from_net_downwelling",
    units="W m-2",
    standard_name="surface_upwelling_longwave_flux",
    long_name="Upwelling longwave flux",
    description="The calculation of upwelling longwave radiative flux from net surface longwave "
    "and downwelling surface longwave fluxes.",
    var_name="rlus",
    compute=indices.longwave_upwelling_radiation_from_net_downwelling,
)



from __future__ import annotations

import os

import numpy as np
import pandas as pd
import pytest
import xarray as xr
from dask import compute

from xclim.core.options import set_options
from xclim.indices import run_length as rl
from xclim.testing.helpers import assert_lazy

K2C = 273.15


class TestSuspiciousRun:
    def test_simple(self, tas_series):
        t = np.zeros(365)
        tas = tas_series(t, start="2000-01-01")
        sus = rl.suspicious_run(tas)
        # Only zeroes
        assert sus.all()

        t = np.zeros(365)
        t[30:39] = 5
        tas = tas_series(t, start="2000-01-01")
        sus = rl.suspicious_run(tas, thresh=0)
        # Not enough 5s to trigger suspicion
        assert not sus[30:39].all()
        assert not sus[0:10].all()

        t = np.zeros(365)
        t[30:40] = 1
        tas = tas_series(t, start="2000-01-01")
        sus = rl.suspicious_run(tas, thresh=0)
        # Run of 10 identical values
        assert sus[30:40].all()
        assert not sus[30:41].all()

    def test_above_thresh(self, tas_series):
        t = np.zeros(365)
        t[30:40] = 0.1
        t[40:50] = 1e-6
        t[50:60] = 0.0001
        t[60:65] = 1e-9
        tas = tas_series(t, start="2000-01-01")

        sus = rl.suspicious_run(tas, thresh=0, window=5)
        assert not sus[:30].any()
        assert sus[30:65].all()
        assert not sus[65:].any()

        sus = rl.suspicious_run(tas, thresh=1e-9, window=5)
        assert sus[30:60].all()
        assert not sus[60:].any()

        sus = rl.suspicious_run(tas, thresh=1e-5, window=5)
        assert sus[30:40].all()
        assert not sus[40:50].any()
        assert sus[50:60].all()
        assert not sus[60:].any()

        sus = rl.suspicious_run(tas, thresh=0, window=11)
        assert not sus.any()

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_dask(self, use_dask):
        values = np.zeros((10, 200))
        time = pd.date_range("2015-01-01", periods=200, freq="D")
        values[:, :10] = 1
        values[9, :] = 1
        da = xr.DataArray(values, coords={"time": time}, dims=("qq", "time"))

        if use_dask:
            da = da.chunk({"qq": 2})

        sus = rl.suspicious_run(da, thresh=0)
        assert sus[:, :10].all()
        assert not sus[1, 10:].any()
        assert sus[9].all()

        sus = rl.suspicious_run(da)
        assert sus.all()


@pytest.fixture(scope="module", params=[True, False], autouse=True)
def ufunc(request):
    with set_options(run_length_ufunc=request.param):
        yield request.param


@pytest.mark.parametrize("use_dask", [True, False])
@pytest.mark.parametrize("index", ["first", "last"])
def test_rle(ufunc, use_dask, index):
    if use_dask and ufunc:
        pytest.skip("rle_1d is not implemented for dask arrays.")

    values = np.zeros((10, 365, 4, 4))
    time = pd.date_range("2000-01-01", periods=365, freq="D")
    values[:, 1:11, ...] = 1
    da = xr.DataArray(values, coords={"time": time}, dims=("a", "time", "b", "c"))

    if ufunc:
        da = da[0, :, 0, 0]
        v, l, p = rl.rle_1d(da != 0)  # noqa: E741
        np.testing.assert_array_equal(v, [False, True, False])
        np.testing.assert_array_equal(l, [1, 10, 354])
        np.testing.assert_array_equal(p, [0, 1, 11])
    else:
        if use_dask:
            da = da.chunk({"a": 1, "b": 2})

        out = rl.rle(da != 0, index=index).mean(["a", "b", "c"])
        if index == "last":
            expected = np.zeros(365)
            expected[1:10] = np.nan
            expected[10] = 10
        else:
            expected = np.zeros(365)
            expected[1] = 10
            expected[2:11] = np.nan
        np.testing.assert_array_equal(out, expected)


@pytest.mark.parametrize("use_dask", [True, False])
@pytest.mark.parametrize("index", ["first", "last"])
def test_runs_with_holes_identity(use_dask, index):
    # This test reproduces the behaviour or `rle`
    values = np.zeros((10, 365, 4, 4))
    time = pd.date_range("2000-01-01", periods=365, freq="D")
    values[:, 1:11, ...] = 1
    da = xr.DataArray(values, coords={"time": time}, dims=("a", "time", "b", "c"))

    if use_dask:
        da = da.chunk({"a": 1, "b": 2})

    events = rl.runs_with_holes(da != 0, 1, da == 0, 1)
    expected = da
    np.testing.assert_array_equal(events, expected)


def test_runs_with_holes():
    values = np.zeros(365)
    time = pd.date_range("2000-01-01", periods=365, freq="D")
    a = [0, 1, 0, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0]
    values[0 : len(a)] = a
    da = xr.DataArray(values, coords={"time": time}, dims=("time"))

    events = rl.runs_with_holes(da == 1, 1, da == 0, 3)

    expected = values * 0
    expected[1:11] = 1
    expected[15:20] = 1

    np.testing.assert_array_equal(events, expected)


class TestStatisticsRun:
    nc_pr = os.path.join("NRCANdaily", "nrcan_canada_daily_pr_1990.nc")

    def test_simple(self):
        values = np.zeros(365)
        time = pd.date_range("7/1/2000", periods=len(values), freq="D")
        values[1:11] = 1
        da = xr.DataArray(values != 0, coords={"time": time}, dims="time")
        lt = da.resample(time="ME").map(rl.rle_statistics, reducer="max", window=1)
        assert lt[0] == 10
        np.testing.assert_array_equal(lt[1:], 0)

        # resample after
        lt = rl.rle_statistics(da, freq="ME", reducer="max", window=1, ufunc_1dim=False)
        assert lt[0] == 10
        np.testing.assert_array_equal(lt[1:], 0)

    def test_start_at_0(self):
        values = np.zeros(365)
        time = pd.date_range("7/1/2000", periods=len(values), freq="D")
        values[0:10] = 1
        da = xr.DataArray(values != 0, coords={"time": time}, dims="time")
        lt = da.resample(time="ME").map(rl.rle_statistics, reducer="max", window=1)
        assert lt[0] == 10
        np.testing.assert_array_equal(lt[1:], 0)

        # resample after
        lt = rl.rle_statistics(da, freq="ME", reducer="max", window=1, ufunc_1dim=False)
        assert lt[0] == 10
        np.testing.assert_array_equal(lt[1:], 0)

    def test_end_start_at_0(self):
        values = np.zeros(365)
        time = pd.date_range("7/1/2000", periods=len(values), freq="D")
        values[-10:] = 1
        da = xr.DataArray(values != 0, coords={"time": time}, dims="time")

        lt = da.resample(time="ME").map(rl.rle_statistics, reducer="max", window=1)
        assert lt[-1] == 10
        np.testing.assert_array_equal(lt[:-1], 0)

        # resample after
        lt = rl.rle_statistics(da, freq="ME", reducer="max", window=1, ufunc_1dim=False)
        assert lt[-1] == 10
        np.testing.assert_array_equal(lt[:-1], 0)

    def test_all_true(self):
        values = np.ones(365)
        time = pd.date_range("7/1/2000", periods=len(values), freq="D")
        da = xr.DataArray(values != 0, coords={"time": time}, dims="time")

        lt = da.resample(time="ME").map(rl.rle_statistics, reducer="max", window=1)
        np.testing.assert_array_equal(lt, da.resample(time="ME").count(dim="time"))

        # resample after
        lt = rl.rle_statistics(da, freq="ME", reducer="max", window=1, ufunc_1dim=False)
        expected = np.zeros(12)
        expected[0] = 365
        np.testing.assert_array_equal(lt, expected)
        assert (lt.values == expected).any()

    def test_almost_all_true(self):
        values = np.ones(365)
        values[35] = 0
        time = pd.date_range("7/1/2000", periods=len(values), freq="D")
        da = xr.DataArray(values != 0, coords={"time": time}, dims="time")

        lt = da.resample(time="ME").map(rl.rle_statistics, reducer="max", window=1)
        n = da.resample(time="ME").count(dim="time")
        np.testing.assert_array_equal(lt[0], n[0])
        np.testing.assert_array_equal(lt[1], 26)

        # resample after
        lt = rl.rle_statistics(da, freq="ME", reducer="max", window=1, ufunc_1dim=False)
        expected = np.zeros(12)
        expected[0], expected[1] = 35, 365 - 35 - 1
        np.testing.assert_array_equal(lt[0], expected[0])
        np.testing.assert_array_equal(lt[1], expected[1])

    def test_other_stats(self):
        values = np.ones(365)
        values[35] = 0
        time = pd.date_range("1/1/2000", periods=len(values), freq="D")
        da = xr.DataArray(values != 0, coords={"time": time}, dims="time")

        lt = da.resample(time="YS").map(rl.rle_statistics, reducer="min", window=1)
        assert lt == 35

        lt = da.resample(time="YS").map(rl.rle_statistics, reducer="mean", window=36)
        assert lt == 329

        lt = da.resample(time="YS").map(rl.rle_statistics, reducer="std", window=1)
        assert lt == 147

        # resample after
        lt = rl.rle_statistics(da, freq="YS", reducer="min", window=1, ufunc_1dim=False)
        assert lt == 35

        lt = rl.rle_statistics(
            da, freq="YS", reducer="mean", window=36, ufunc_1dim=False
        )
        assert lt == 329

        lt = rl.rle_statistics(da, freq="YS", reducer="std", window=1, ufunc_1dim=False)
        assert lt == 147

    @pytest.mark.parametrize("op", ["min", "max"])
    def test_resampling_order(self, op):
        values = np.ones(365)
        values[35:45] = 0
        time = pd.date_range("1/1/2000", periods=len(values), freq="D")
        da = xr.DataArray(values != 0, coords={"time": time}, dims="time")
        lt_resample_before = da.resample(time="MS").map(
            rl.rle_statistics, reducer=op, window=1, ufunc_1dim=False
        )
        lt_resample_after = rl.rle_statistics(
            da, freq="MS", reducer=op, window=1, ufunc_1dim=False
        )
        assert (lt_resample_before != lt_resample_after).any()

        values = np.zeros(365)
        values[0:-1:31] = 1
        time = pd.date_range("1/1/2000", periods=len(values), freq="D")
        da = xr.DataArray(values != 0, coords={"time": time}, dims="time")
        lt_resample_before = da.resample(time="MS").map(
            rl.rle_statistics, reducer=op, window=1, ufunc_1dim=False
        )
        lt_resample_after = rl.rle_statistics(
            da, freq="MS", reducer=op, window=1, ufunc_1dim=False
        )
        assert (lt_resample_before == lt_resample_after).any()


class TestFirstRun:
    nc_pr = os.path.join("NRCANdaily", "nrcan_canada_daily_pr_1990.nc")

    def test_real_simple(self):
        a = xr.DataArray(np.zeros(100, bool), dims=("x",))
        a[10:20] = 1
        i = rl.first_run(a, 5, dim="x")
        assert 10 == i

    def test_real_data(self, open_dataset):
        # FIXME: No test here?!
        # n-dim version versus ufunc
        da3d = open_dataset(self.nc_pr).pr[:, 40:50, 50:68] != 0
        da3d.resample(time="ME").map(rl.first_run, window=5)

    @pytest.mark.parametrize(
        "coord,expected",
        [(False, 30), (True, np.datetime64("2000-01-31")), ("dayofyear", 31)],
    )
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_simple(self, tas_series, coord, expected, use_dask, ufunc):
        # if use_dask and ufunc:
        #     pytest.xfail("Ufunc run length algorithms not implemented for dask arrays.")
        t = np.zeros(60)
        t[30:40] = 2
        tas = tas_series(t, start="2000-01-01")
        runs = xr.concat((tas, tas), dim="dim0")

        if use_dask:
            runs = runs.chunk({"time": -1 if ufunc else 10, "dim0": 1})

        out = rl.first_run(runs, window=1, dim="time", coord=coord)
        np.testing.assert_array_equal(out.load(), expected)

    @pytest.mark.parametrize("use_dask", [True, False])
    @pytest.mark.parametrize(
        "coord,expected",
        [
            (False, [0, 0]),
            (True, [np.datetime64("2000-01-01"), np.datetime64("2000-02-01")]),
            ("dayofyear", [1, 32]),
        ],
    )
    def test_resample_after(self, tas_series, coord, expected, use_dask):
        t = np.zeros(60)
        t[0] = 2
        t[30:40] = 2
        tas = tas_series(t, start="2000-01-01")
        runs = xr.concat((tas, tas), dim="dim0")

        if use_dask:
            runs = runs.chunk({"time": -1 if ufunc else 10, "dim0": 1})

        out = rl.first_run(
            runs, window=1, dim="time", coord=coord, freq="MS", ufunc_1dim=False
        )
        np.testing.assert_array_equal(out.load(), np.array([expected, expected]))


class TestWindowedRunEvents:
    @pytest.mark.parametrize("index", ["first", "last"])
    def test_simple(self, index):
        a = xr.DataArray(np.zeros(50, bool), dims=("x",))
        a[4:7] = True
        a[34:45] = True
        assert rl.windowed_run_events(a, 3, dim="x", index=index) == 2


class TestWindowedRunCount:
    @pytest.mark.parametrize("index", ["first", "last"])
    def test_simple(self, index):
        a = xr.DataArray(np.zeros(50, bool), dims=("time",))
        a[4:7] = True
        a[34:45] = True
        assert rl.windowed_run_count(a, 3, dim="time", index=index) == len(
            a[4:7]
        ) + len(a[34:45])


class TestWindowedMaxRunSum:
    @pytest.mark.parametrize("index", ["first", "last"])
    def test_simple(self, index):
        a = xr.DataArray(np.zeros(50, float), dims=("time",))
        a[4:6] = 5  # too short
        a[25:30] = 5  # long enough, but not max
        a[35:45] = 5  # max sum => yields 10*5
        assert rl.windowed_max_run_sum(a, 3, dim="time", index=index) == 50


class TestLastRun:
    @pytest.mark.parametrize(
        "coord,expected",
        [(False, 39), (True, np.datetime64("2000-02-09")), ("dayofyear", 40)],
    )
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_simple(self, tas_series, coord, expected, use_dask, ufunc):
        # if use_dask and ufunc:
        #     pytest.xfail("Ufunc run length algorithms not implemented for dask arrays.")
        t = np.zeros(60)
        t[30:40] = 2
        tas = tas_series(t, start="2000-01-01")
        runs = xr.concat((tas, tas), dim="dim0")

        if use_dask:
            runs = runs.chunk({"time": -1 if ufunc else 10, "dim0": 1})

        out = rl.last_run(runs, window=1, dim="time", coord=coord, ufunc_1dim=ufunc)
        np.testing.assert_array_equal(out.load(), expected)

    @pytest.mark.parametrize("use_dask", [True, False])
    @pytest.mark.parametrize(
        "coord,expected",
        [
            (False, [30, 8]),
            (True, [np.datetime64("2000-01-31"), np.datetime64("2000-02-09")]),
            ("dayofyear", [31, 40]),
        ],
    )
    def test_resample_after(self, tas_series, coord, expected, use_dask):
        t = np.zeros(60)
        t[0] = 2
        t[30:40] = 2
        tas = tas_series(t, start="2000-01-01")
        runs = xr.concat((tas, tas), dim="dim0")

        if use_dask:
            runs = runs.chunk({"time": -1 if ufunc else 10, "dim0": 1})

        out = rl.last_run(
            runs, window=1, dim="time", coord=coord, freq="MS", ufunc_1dim=False
        )
        np.testing.assert_array_equal(out.load(), np.array([expected, expected]))


def test_run_bounds_synthetic():
    run = xr.DataArray(
        [0, 1, 1, 1, 0, 0, 1, 1, 1, 0], dims="x", coords={"x": np.arange(10) ** 2}
    )
    bounds = rl.run_bounds(run, "x", coord=True)
    np.testing.assert_array_equal(bounds, [[1, 36], [16, 81]])

    bounds = rl.run_bounds(run, "x", coord=False)
    np.testing.assert_array_equal(bounds, [[1, 6], [4, 9]])


def test_run_bounds_data(open_dataset):
    era5 = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")
    cond = era5.tas.rolling(time=7).mean() > 285

    bounds = rl.run_bounds(cond, "time")  # def coord = True
    np.testing.assert_array_equal(
        bounds.isel(location=0, events=0),
        pd.to_datetime(["1990-06-21", "1990-10-26"]).values,
    )

    bounds = rl.run_bounds(cond, "time", coord="dayofyear")
    np.testing.assert_array_equal(bounds.isel(location=1, events=4), [279, 283])
    assert bounds.events.size == 15


def test_keep_longest_run_synthetic():
    runs = xr.DataArray([0, 1, 1, 1, 0, 0, 1, 1, 1, 0], dims="time").astype(bool)
    lrun = rl.keep_longest_run(runs, "time")
    np.testing.assert_array_equal(
        lrun, np.array([0, 1, 1, 1, 0, 0, 0, 0, 0, 0], dtype=bool)
    )


def test_keep_longest_run_data(open_dataset):
    era5 = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")
    cond = era5.swe > 0.002
    lrun = rl.keep_longest_run(cond, "time")
    np.testing.assert_array_equal(
        lrun.isel(time=slice(651, 658), location=2),
        np.array([0, 0, 0, 1, 1, 1, 1], dtype=bool),
    )

    xr.testing.assert_equal(
        rl.keep_longest_run(cond, "time").sum("time"),
        rl.longest_run(cond, "time"),
    )


class TestRunsWithDates:
    @pytest.mark.parametrize(
        "date,end,expected",
        [
            ("07-01", 210, 70),
            ("07-01", 190, 50),
            ("04-01", 150, 0),  # date falls early
            ("11-01", 150, 165),  # date ends late
            (None, 150, 10),  # no date, real length
        ],
    )
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_season_length(self, tas_series, date, end, expected, use_dask, ufunc):
        # if use_dask and ufunc:
        #     pytest.xfail("Ufunc run length algorithms not implemented for dask arrays.")
        t = np.zeros(360)
        t[140:end] = 1
        tas = tas_series(t, start="2000-01-01")
        runs = xr.concat((tas, tas), dim="dim0")
        runs = runs == 1

        if use_dask:
            runs = runs.chunk({"time": -1 if ufunc else 10, "dim0": 1})

        out = rl.season_length(
            runs,
            window=1,
            dim="time",
            mid_date=date,
        )
        np.testing.assert_array_equal(np.mean(out.load()), expected)

    @pytest.mark.parametrize(
        "coord,date,end,expected",
        [
            ("dayofyear", "07-01", 210, 211),
            (False, "07-01", 190, 190),
            ("dayofyear", "04-01", 150, np.nan),  # date falls early
            ("dayofyear", "11-01", 150, 306),  # date ends late
        ],
    )
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_run_end_after_date(
        self, tas_series, coord, date, end, expected, use_dask, ufunc
    ):
        # if use_dask and ufunc:
        #     pytest.xfail("Ufunc run length algorithms not implemented for dask arrays.")

        t = np.zeros(360)
        t[140:end] = 1
        tas = tas_series(t, start="2000-01-01")
        runs = xr.concat((tas, tas), dim="dim0")
        runs = runs == 1

        if use_dask:
            runs = runs.chunk({"time": -1 if ufunc else 10, "dim0": 1})

        out = rl.run_end_after_date(runs, window=1, date=date, dim="time", coord=coord)
        np.testing.assert_array_equal(np.mean(out.load()), expected)

    @pytest.mark.parametrize(
        "coord,date,beg,expected",
        [
            ("dayofyear", "07-01", 210, 211),
            (False, "07-01", 190, 190),
            ("dayofyear", "04-01", False, np.nan),  # no run
            ("dayofyear", "11-01", 150, 306),  # run already started
        ],
    )
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_first_run_after_date(
        self, tas_series, coord, date, beg, expected, use_dask, ufunc
    ):
        # if use_dask and ufunc:
        #     pytest.xfail("Ufunc run length algorithms not implemented for dask arrays.")
        t = np.zeros(365)
        if beg:
            t[beg:] = 1
        tas = tas_series(t, start="2000-01-01")
        runs = xr.concat((tas, tas), dim="dim0")
        runs = runs == 1

        if use_dask:
            runs = runs.chunk({"time": -1 if ufunc else 10, "dim0": 1})

        out = rl.first_run_after_date(
            runs, window=1, date=date, dim="time", coord=coord
        )
        np.testing.assert_array_equal(np.mean(out.load()), expected)

    @pytest.mark.parametrize(
        "coord,date,end,expected",
        [
            ("dayofyear", "07-01", 210, 183),
            (False, "07-01", 190, 182),
            ("dayofyear", "04-01", 150, np.nan),  # date falls early
            ("dayofyear", "11-01", 150, 150),  # date ends late
        ],
    )
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_last_run_before_date(
        self, tas_series, coord, date, end, expected, use_dask, ufunc
    ):
        # if use_dask and ufunc:
        #     pytest.xfail("Ufunc run length algorithms not implemented for dask arrays.")
        t = np.zeros(360)
        t[140:end] = 1
        tas = tas_series(t, start="2000-01-01")
        runs = xr.concat((tas, tas), dim="dim0")
        runs = runs == 1

        if use_dask:
            runs = runs.chunk({"time": -1 if ufunc else 10, "dim0": 1})

        out = rl.last_run_before_date(
            runs, window=1, date=date, dim="time", coord=coord
        )
        np.testing.assert_array_equal(np.mean(out.load()), expected)

    @pytest.mark.parametrize(
        "func",
        [rl.last_run_before_date, rl.run_end_after_date],
    )
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_run_with_dates_no_date(self, tas_series, use_dask, func, ufunc):
        # if use_dask and ufunc:
        #     pytest.xfail("Ufunc run length algorithms not implemented for dask arrays.")
        t = np.ones(90)
        tas = tas_series(t, start="2000-01-01")
        runs = xr.concat((tas, tas), dim="dim0")
        runs = runs == 1

        if use_dask:
            runs = runs.chunk({"time": 10, "dim0": 1})

        out = runs.resample(time="MS").map(func, window=1, date="07-01", dim="time")
        assert out.isnull().all()

    @pytest.mark.parametrize(
        "calendar,expected",
        [("standard", [61, 60]), ("365_day", [60, 60]), ("366_day", [61, 61])],
    )
    def test_run_with_dates_different_calendars(self, calendar, expected):
        time = xr.cftime_range(
            "2004-01-01", end="2005-12-31", freq="D", calendar=calendar
        )
        tas = np.zeros(time.size)
        start = np.where((time.day == 1) & (time.month == 3))[0]
        tas[start[0] : start[0] + 250] = 5
        tas[start[1] : start[1] + 250] = 5
        tas = xr.DataArray(tas, coords={"time": time}, dims=("time",))
        out = (
            (tas > 0)
            .resample(time="YS-MAR")
            .map(rl.first_run_after_date, date="03-01", window=2)
        )
        np.testing.assert_array_equal(out.values[1:], expected)

        out = (
            (tas > 0)
            .resample(time="YS-MAR")
            .map(rl.season_length, mid_date="03-02", window=2)
        )
        np.testing.assert_array_equal(out.values[1:], [250, 250])

        out = (
            (tas > 0)
            .resample(time="YS-MAR")
            .map(rl.run_end_after_date, date="03-03", window=2)
        )
        np.testing.assert_array_equal(out.values[1:], np.array(expected) + 250)

        out = (
            (tas > 0)
            .resample(time="YS-MAR")
            .map(rl.last_run_before_date, date="03-02", window=2)
        )
        np.testing.assert_array_equal(out.values[1:], np.array(expected) + 1)

    @pytest.mark.parametrize("func", [rl.first_run_after_date, rl.run_end_after_date])
    def test_too_many_dates(self, func, tas_series):
        tas = tas_series(np.zeros(730), start="2000-01-01")
        with pytest.raises(ValueError, match="More than 1 instance of date"):
            func((tas == 0), date="03-01", window=5)


@pytest.mark.parametrize("use_dask", [True, False])
def test_lazy_indexing(use_dask):
    idx = xr.DataArray([[0, 10], [33, 99]], dims=("x", "y"))
    idx = idx.assign_coords(x2=idx.x**2)
    da = xr.DataArray(np.arange(100), dims=("time",))
    db = xr.DataArray(-np.arange(100), dims=("time",))

    if use_dask:
        idx = idx.chunk({"x": 1})

    # Ensure tasks are different
    with assert_lazy:
        outa = rl.lazy_indexing(da, idx)
        outb = rl.lazy_indexing(db, idx)
    outa, outb = compute(outa, outb)

    assert set(outa.dims) == {"x", "y"}
    np.testing.assert_array_equal(idx, outa)
    np.testing.assert_array_equal(idx, -outb)
    assert "time" not in outa.coords


@pytest.mark.parametrize("use_dask", [True, False])
def test_lazy_indexing_special_cases(use_dask, random):
    a = xr.DataArray(random.random((10, 10, 10)), dims=("x", "y", "z"))
    b = xr.DataArray(random.random((10, 10, 10)), dims=("x", "y", "z"))

    if use_dask:
        a = a.chunk({"y": 5, "z": 5})
        b = b.chunk({"y": 5, "z": 1})

    with pytest.raises(ValueError):
        rl.lazy_indexing(a, b)

    b = b.argmin("x").argmin("y")

    with pytest.raises(ValueError, match="more than one dimension more than index"):
        rl.lazy_indexing(a, b)

    c = xr.DataArray([1], dims=("x",)).chunk()[0]
    b["z"] = np.arange(b.z.size)
    rl.lazy_indexing(b, c)


@pytest.mark.parametrize("use_dask", [True, False])
def test_season(use_dask, tas_series, ufunc):
    # if use_dask and ufunc:
    #     pytest.xfail("Ufunc run length algorithms not implemented for dask arrays.")
    t = np.zeros(360)
    t[140:150] = 1
    tas = tas_series(t, start="2000-01-01")
    runs = xr.concat((tas, tas), dim="dim0")
    runs = runs >= 1

    if use_dask:
        runs = runs.chunk({"time": -1 if ufunc else 10, "dim0": 1})

    out = rl.season(runs, window=2)
    np.testing.assert_array_equal(out.start.load(), [140, 140])
    np.testing.assert_array_equal(out.end.load(), [150, 150])
    np.testing.assert_array_equal(out.length.load(), [10, 10])



# pylint: disable=no-member,missing-kwoa
"""
SDBA Diagnostic Testing Module
==============================

This module is meant to compare results with those expected from papers, or create figures illustrating the
behavior of sdba methods and utilities.
"""
from __future__ import annotations

import warnings

import numpy as np
from scipy.stats import gaussian_kde, scoreatpercentile

from xclim.sdba.adjustment import (
    DetrendedQuantileMapping,
    EmpiricalQuantileMapping,
    QuantileDeltaMapping,
)
from xclim.sdba.processing import adapt_freq
from xclim.testing.sdba_utils import cannon_2015_rvs, series

try:
    from matplotlib import pyplot as plt
except ModuleNotFoundError:
    warnings.warn("Matplotlib not found, plot-generating functions will not work.")
    plt = None


__all__ = ["adapt_freq_graph", "cannon_2015_figure_2", "synth_rainfall"]


def synth_rainfall(shape, scale=1, wet_freq=0.25, size=1):
    r"""Return gamma distributed rainfall values for wet days.

    Notes
    -----
    The probability density for the Gamma distribution is:

    .. math::

        p(x) = x^{k-1}\frac{e^{-x/\theta}}{\theta^k\Gamma(k)},

    where :math:`k` is the shape and :math:`\theta` the scale, and :math:`\Gamma` is the Gamma function.
    """
    is_wet = np.random.binomial(1, p=wet_freq, size=size)
    wet_intensity = np.random.gamma(shape, scale, size)
    return np.where(is_wet, wet_intensity, 0)


def cannon_2015_figure_2():
    """Create a graphic similar to figure 2 of Cannon et al. 2015."""
    # noqa: D103
    if plt is None:
        raise ModuleNotFoundError("Matplotlib not found.")

    n = 10000
    ref, hist, sim = cannon_2015_rvs(n, random=False)
    QM = EmpiricalQuantileMapping(kind="*", group="time", interp="linear")
    QM.train(ref, hist)
    sim_eqm = QM.predict(sim)

    DQM = DetrendedQuantileMapping(kind="*", group="time", interp="linear")
    DQM.train(ref, hist)
    sim_dqm = DQM.predict(sim, degree=0)

    QDM = QuantileDeltaMapping(kind="*", group="time", interp="linear")
    QDM.train(ref, hist)
    sim_qdm = QDM.predict(sim)

    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(11, 4))
    x = np.linspace(0, 105, 50)
    ax1.plot(x, gaussian_kde(ref)(x), color="r", label="Obs hist")
    ax1.plot(x, gaussian_kde(hist)(x), color="k", label="GCM hist")
    ax1.plot(x, gaussian_kde(sim)(x), color="blue", label="GCM simure")

    ax1.plot(x, gaussian_kde(sim_qdm)(x), color="lime", label="QDM future")
    ax1.plot(x, gaussian_kde(sim_eqm)(x), color="darkgreen", ls="--", label="QM future")
    ax1.plot(x, gaussian_kde(sim_dqm)(x), color="lime", ls=":", label="DQM future")
    ax1.legend(frameon=False)
    ax1.set_xlabel("Value")
    ax1.set_ylabel("Density")

    tau = np.array([0.25, 0.5, 0.75, 0.95, 0.99]) * 100
    bc_gcm = (
        scoreatpercentile(sim, tau) - scoreatpercentile(hist, tau)
    ) / scoreatpercentile(hist, tau)
    bc_qdm = (
        scoreatpercentile(sim_qdm, tau) - scoreatpercentile(ref, tau)
    ) / scoreatpercentile(ref, tau)
    bc_eqm = (
        scoreatpercentile(sim_eqm, tau) - scoreatpercentile(ref, tau)
    ) / scoreatpercentile(ref, tau)
    bc_dqm = (
        scoreatpercentile(sim_dqm, tau) - scoreatpercentile(ref, tau)
    ) / scoreatpercentile(ref, tau)

    ax2.plot([0, 1], [0, 1], ls=":", color="blue")
    ax2.plot(bc_gcm, bc_gcm, "-", color="blue", label="GCM")
    ax2.plot(bc_gcm, bc_qdm, marker="o", mfc="lime", label="QDM")
    ax2.plot(
        bc_gcm,
        bc_eqm,
        marker="o",
        mfc="darkgreen",
        ls=":",
        color="darkgreen",
        label="QM",
    )
    ax2.plot(
        bc_gcm,
        bc_dqm,
        marker="s",
        mec="lime",
        mfc="w",
        ls="--",
        color="lime",
        label="DQM",
    )

    for i, s in enumerate(tau / 100):
        ax2.text(bc_gcm[i], bc_eqm[i], f"{s}  ", ha="right", va="center", fontsize=9)
    ax2.set_xlabel("GCM relative change")
    ax2.set_ylabel("Bias adjusted relative change")
    ax2.legend(loc="upper left", frameon=False)
    ax2.set_aspect("equal")
    plt.tight_layout()
    return fig


def adapt_freq_graph():
    """Create a graphic with the additive adjustment factors estimated after applying the adapt_freq method."""
    if plt is None:
        raise ModuleNotFoundError("Matplotlib not found.")

    n = 10000
    x = series(synth_rainfall(2, 2, wet_freq=0.25, size=n), "pr")  # sim
    y = series(synth_rainfall(2, 2, wet_freq=0.5, size=n), "pr")  # ref

    xp = adapt_freq(x, y, thresh=0).sim_ad  # noqa

    fig, (ax1, ax2) = plt.subplots(2, 1)
    sx = x.sortby(x)
    sy = y.sortby(y)
    sxp = xp.sortby(xp)

    # Original and corrected series
    ax1.plot(sx.values, color="blue", lw=1.5, label="x : sim")
    ax1.plot(sxp.values, color="pink", label="xp : sim corrected")
    ax1.plot(sy.values, color="k", label="y : ref")
    ax1.legend()

    # Compute qm factors
    qm_add = QuantileDeltaMapping(kind="+", group="time").train(y, x).ds
    qm_mul = QuantileDeltaMapping(kind="*", group="time").train(y, x).ds

    qm_add_p = QuantileDeltaMapping(kind="+", group="time").train(y, xp).ds
    qm_mul_p = QuantileDeltaMapping(kind="*", group="time").train(y, xp).ds

    qm_add.cf.plot(ax=ax2, color="cyan", ls="--", label="+: y-x")
    qm_add_p.cf.plot(ax=ax2, color="cyan", label="+: y-xp")

    qm_mul.cf.plot(ax=ax2, color="brown", ls="--", label="*: y/x")
    qm_mul_p.cf.plot(ax=ax2, color="brown", label="*: y/xp")

    ax2.legend(loc="upper left", frameon=False)
    return fig



from __future__ import annotations

from pathlib import Path


def _find_current_folder():
    """Find the folder containing the notebooks.

    Needed in order to run the notebooks from the docs/notebooks folder.
    """
    return Path(__file__).absolute().parent



from __future__ import annotations

import numpy as np

from xclim import atmos


class TestWindSpeedIndicators:
    def test_calm_windy_days(self, open_dataset):
        ds = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")
        sfcwind, _ = atmos.wind_speed_from_vector(
            ds.uas, ds.vas, calm_wind_thresh="0 m/s"
        )

        calm = atmos.calm_days(sfcwind, thresh="5 m/s")
        windy = atmos.windy_days(sfcwind, thresh="5 m/s")
        c = sfcwind.resample(time="MS").count()
        np.testing.assert_array_equal(calm + windy, c)


class TestSfcWindMax:
    def test_sfcWind_max(self, open_dataset):
        ds = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")
        sfcWind, _ = atmos.wind_speed_from_vector(ds.uas, ds.vas)

        sfcWind_max = atmos.sfcWind_max(sfcWind)
        c = sfcWind.resample(time="YS").max()
        np.testing.assert_array_equal(sfcWind_max, c)


class TestSfcWindMean:
    def test_sfcWind_mean(self, open_dataset):
        ds = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")
        sfcWind, _ = atmos.wind_speed_from_vector(ds.uas, ds.vas)

        sfcWind_mean = atmos.sfcWind_mean(sfcWind)
        c = sfcWind.resample(time="YS").mean()
        np.testing.assert_array_equal(sfcWind_mean, c)


class TestSfcWindMin:
    def test_sfcWind_min(self, open_dataset):
        ds = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")
        sfcWind, _ = atmos.wind_speed_from_vector(ds.uas, ds.vas)

        sfcWind_min = atmos.sfcWind_min(sfcWind)
        c = sfcWind.resample(time="YS").min()
        np.testing.assert_array_equal(sfcWind_min, c)


class TestSfcWindmaxMax:
    def test_sfcWindmax_max(self, open_dataset):
        ds = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")
        sfcWindmax, _ = atmos.wind_speed_from_vector(ds.uas, ds.vas)

        sfcWindmax_max = atmos.sfcWindmax_max(sfcWindmax)
        c = sfcWindmax.resample(time="YS").max()
        np.testing.assert_array_equal(sfcWindmax_max, c)


class TestSfcWindmaxMean:
    def test_sfcWindmax_mean(self, open_dataset):
        ds = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")
        sfcWindmax, _ = atmos.wind_speed_from_vector(ds.uas, ds.vas)

        sfcWindmax_mean = atmos.sfcWindmax_mean(sfcWindmax)
        c = sfcWindmax.resample(time="YS").mean()
        np.testing.assert_array_equal(sfcWindmax_mean, c)


class TestSfcWindmaxMin:
    def test_sfcWindmax_mean(self, open_dataset):
        ds = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")
        sfcWindmax, _ = atmos.wind_speed_from_vector(ds.uas, ds.vas)

        sfcWindmax_min = atmos.sfcWindmax_min(sfcWindmax)
        c = sfcWindmax.resample(time="YS").min()
        np.testing.assert_array_equal(sfcWindmax_min, c)



# This File is for testing atmos indicators that fit neither in test_precip or test_temperature
# Expected values might be the same as for the indices tests, see test_indices comments.
from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

from xclim import atmos, set_options
from xclim.indices.helpers import make_hourly_temperature

K2C = 273.16


def test_wind_speed_from_vectors():
    uas = xr.DataArray(np.array([3.0, -3.0]), dims=["x"])
    uas.attrs["units"] = "m s-1"
    vas = xr.DataArray(np.array([4.0, -4.0]), dims=["x"])
    vas.attrs["units"] = "m s-1"

    wind, winddir = atmos.wind_speed_from_vector(uas=uas, vas=vas)
    np.testing.assert_allclose(wind, [5.0, 5.0])
    np.testing.assert_allclose(winddir, [216.86989764584402, 36.86989764584402])

    # missing values
    uas[0] = np.nan
    wind, winddir = atmos.wind_speed_from_vector(uas=uas, vas=vas)
    np.testing.assert_array_equal(wind.isnull(), [True, False])
    np.testing.assert_array_equal(winddir.isnull(), [True, False])

    # Calm thresh and northerly
    uas[:] = 0
    vas[0] = 0.9
    vas[1] = -1.1
    wind, winddir = atmos.wind_speed_from_vector(
        uas=uas, vas=vas, calm_wind_thresh="1 m/s"
    )
    np.testing.assert_array_equal(wind, [0.9, 1.1])
    np.testing.assert_allclose(winddir, [0.0, 360.0])


def test_wind_vector_from_speed():
    sfcWind = xr.DataArray(np.array([3.0, 5.0, 0.2]), dims=["x"])
    sfcWind.attrs["units"] = "m s-1"
    sfcWindfromdir = xr.DataArray(np.array([360.0, 36.86989764584402, 0.0]), dims=["x"])
    sfcWindfromdir.attrs["units"] = "degree"

    uas, vas = atmos.wind_vector_from_speed(
        sfcWind=sfcWind, sfcWindfromdir=sfcWindfromdir
    )
    np.testing.assert_allclose(uas, [0.0, -3.0, 0.0], atol=1e-14)
    np.testing.assert_allclose(vas, [-3.0, -4.0, -0.2], atol=1e-14)

    # missing values
    sfcWind[0] = np.nan
    sfcWindfromdir[1] = np.nan
    uas, vas = atmos.wind_vector_from_speed(
        sfcWind=sfcWind, sfcWindfromdir=sfcWindfromdir
    )
    np.testing.assert_array_equal(uas.isnull(), [True, True, False])
    np.testing.assert_array_equal(vas.isnull(), [True, True, False])


def test_relative_humidity_dewpoint(timeseries):
    np.testing.assert_allclose(
        atmos.relative_humidity_from_dewpoint(
            tas=timeseries(
                np.array([-20, -10, -1, 10, 20, 25, 30, 40, 60]) + K2C, "tas"
            ),
            tdps=timeseries(
                np.array([-15, -10, -2, 5, 10, 20, 29, 20, 30]) + K2C, "tdps"
            ),
        ),
        timeseries([np.nan, 100, 93, 71, 52, 73, 94, 31, 20], "hurs"),
        rtol=0.02,
        atol=1,
    )


def test_humidex(tas_series):
    tas = tas_series([15, 25, 35, 40])
    tas.attrs["units"] = "C"

    dtas = tas_series([10, 15, 25, 25])
    dtas.attrs["units"] = "C"

    # expected values from https://en.wikipedia.org/wiki/Humidex
    h = atmos.humidex(tas, dtas)
    np.testing.assert_array_almost_equal(h, [16, 29, 47, 52], 0)
    assert h.name == "humidex"

    # Test with dataset (#1432)
    ds = xr.Dataset({"tas": tas, "dewpoint": dtas})
    h2 = atmos.humidex(ds=ds, tdps="dewpoint")
    np.testing.assert_array_almost_equal(h, h2)


def test_heat_index(atmosds):
    # Keep just Montreal values for summer as we need tas > 20 degC
    tas = atmosds.tasmax[1][150:170]
    hurs = atmosds.hurs[1][150:170]

    expected = np.array(
        [
            25.0,
            27.0,
            29.0,
            27.0,
            24.0,
            np.nan,
            np.nan,
            23.0,
            24.0,
            np.nan,
            np.nan,
            24.0,
            28.0,
            25.0,
            30.0,
            26.0,
            31.0,
            33.0,
            34.0,
            28.0,
        ]
    )

    with set_options(cf_compliance="raise"):
        hi = atmos.heat_index(tas, hurs)
    np.testing.assert_array_almost_equal(hi, expected, 0)
    assert hi.name == "heat_index"


def test_saturation_vapor_pressure(tas_series):
    tas = tas_series(np.array([-20, -10, -1, 10, 20, 25, 30, 40, 60]) + K2C)
    e_sat_exp = [103, 260, 563, 1228, 2339, 3169, 4247, 7385, 19947]
    e_sat = atmos.saturation_vapor_pressure(
        tas=tas,
        method="sonntag90",
        ice_thresh="0 degC",
    )
    np.testing.assert_allclose(e_sat, e_sat_exp, atol=0.5, rtol=0.005)
    assert e_sat.name == "e_sat"


def test_relative_humidity(tas_series, hurs_series, huss_series, ps_series):
    tas = tas_series(np.array([-10, -10, 10, 20, 35, 50, 75, 95]) + K2C)
    hurs_exp = hurs_series([np.nan, 63.0, 66.0, 34.0, 14.0, 6.0, 1.0, 0.0])
    ps = ps_series([101325] * 8)
    huss = huss_series([0.003, 0.001] + [0.005] * 7)

    hurs = atmos.relative_humidity(
        tas=tas,
        huss=huss,
        ps=ps,
        method="sonntag90",
        ice_thresh="0 degC",
    )
    np.testing.assert_allclose(hurs, hurs_exp, atol=0.5, rtol=0.005)
    assert hurs.name == "hurs"


def test_specific_humidity(tas_series, hurs_series, huss_series, ps_series):
    tas = tas_series(np.array([20, -10, 10, 20, 35, 50, 75, 95]) + K2C)
    hurs = hurs_series([150, 10, 90, 20, 80, 50, 70, 40, 30])
    ps = ps_series(1000 * np.array([100] * 4 + [101] * 4))
    huss_exp = huss_series(
        [np.nan, 1.6e-4, 6.9e-3, 3.0e-3, 2.9e-2, 4.1e-2, 2.1e-1, 5.7e-1]
    )

    huss = atmos.specific_humidity(
        tas=tas,
        hurs=hurs,
        ps=ps,
        method="sonntag90",
        ice_thresh="0 degC",
    )
    np.testing.assert_allclose(huss, huss_exp, atol=1e-4, rtol=0.05)
    assert huss.name == "huss"


def test_specific_humidity_from_dewpoint(tas_series, ps_series, huss_series):
    tdps = tas_series([272, 283, 293])
    ps = ps_series([100000, 105000, 110000])
    # Computed from MetPy
    # >>> from metpy.units import units as u
    # >>> from metpy.calc import specific_humidity_from_dewpoint as sh
    # >>> sh([100000, 105000, 110000] * u.Pa, [272, 283, 293] * u.degK)
    # array([0.0035031, 0.00722795, 0.01319614]) < Unit('dimensionless') >
    huss_exp = huss_series([0.0035031, 0.00722795, 0.01319614])

    huss = atmos.specific_humidity_from_dewpoint(
        tdps=tdps,
        ps=ps,
        method="sonntag90",
    )
    np.testing.assert_allclose(huss, huss_exp, atol=1e-4, rtol=0.05)
    assert huss.name == "huss_fromdewpoint"


def test_snowfall_approximation(pr_series, tasmax_series):
    pr = pr_series(np.ones(10))
    tasmax = tasmax_series(np.arange(10) + K2C)

    prsn = atmos.snowfall_approximation(
        pr, tas=tasmax, thresh="5 degC", method="binary"
    )

    np.testing.assert_allclose(
        prsn, [1, 1, 1, 1, 1, 0, 0, 0, 0, 0], atol=1e-5, rtol=1e-3
    )


def test_rain_approximation(pr_series, tas_series):
    pr = pr_series(np.ones(10))
    tas = tas_series(np.arange(10) + K2C)

    prlp = atmos.rain_approximation(pr, tas=tas, thresh="5 degC", method="binary")

    np.testing.assert_allclose(
        prlp, [0, 0, 0, 0, 0, 1, 1, 1, 1, 1], atol=1e-5, rtol=1e-3
    )


def test_high_precip_low_temp(pr_series, tasmin_series):
    pr = np.zeros(365)
    pr[1:3] = [1, 2]
    pr = pr_series(pr, start="1999-01-01")

    tas = np.zeros(365)
    tas[2:4] = [1, 1]
    tas += K2C
    tas = tasmin_series(tas, start="1999-01-01")

    out = atmos.high_precip_low_temp(
        pr, tas, pr_thresh="1 kg m-2 s-1", tas_thresh="1 C"
    )
    np.testing.assert_array_equal(out, [1])


def test_wind_chill_index(atmosds):
    out = atmos.wind_chill_index(ds=atmosds)

    np.testing.assert_allclose(
        out.isel(time=0), [np.nan, -6.716, -35.617, -8.486, np.nan], rtol=1e-3
    )

    out_us = atmos.wind_chill_index(ds=atmosds, method="US")

    np.testing.assert_allclose(
        out_us.isel(time=0), [-1.429, -6.716, -35.617, -8.486, 2.781], rtol=1e-3
    )


def test_wind_profile(atmosds):
    out = atmos.wind_profile(
        wind_speed=atmosds.sfcWind, h_r="10 m", h="100 m", alpha=1 / 7
    )
    assert out.attrs["units"] == "m s-1"
    assert (out > atmosds.sfcWind).all()


def test_wind_power_potential(atmosds):
    out = atmos.wind_power_potential(wind_speed=atmosds.sfcWind)
    assert out.attrs["units"] == "1"
    assert (out >= 0).all()
    assert (out <= 1).all()


def test_wind_power_potential_from_3h_series():
    """Test a typical computation workflow from 3-hourly time series to daily power production in MWh."""
    from xclim.core.units import convert_units_to
    from xclim.indices.generic import select_resample_op
    from xclim.testing.helpers import test_timeseries

    w = test_timeseries(
        np.ones(96) * 15, variable="sfcWind", start="7/1/2000", units="m s-1", freq="3h"
    )
    out = atmos.wind_power_potential(wind_speed=w)

    # Multiply with nominal capacity
    power = out * 100
    power.attrs["units"] = "MW"
    annual_power = convert_units_to(
        select_resample_op(power, op="integral", freq="D"), "MWh"
    )
    assert (annual_power == 100 * 24).all()


class TestDrynessIndex:
    def test_simple(self, atmosds):
        ds = atmosds.isel(location=3)

        pr = ds.pr
        evspsblpot = ds.evspsblpot

        di = atmos.dryness_index(pr, evspsblpot)
        np.testing.assert_allclose(
            di, np.array([13.355, 102.426, 65.576, 158.078]), rtol=1e-03
        )
        assert di.attrs["long_name"] == "Growing season humidity"

    def test_variable_initial_conditions(self, atmosds):
        ds = atmosds

        pr = ds.pr
        evspsblpot = ds.evspsblpot

        di = atmos.dryness_index(pr, evspsblpot)
        di_wet = atmos.dryness_index(pr, evspsblpot, wo="250 mm")
        di_dry = atmos.dryness_index(pr, evspsblpot, wo="100 mm")

        assert np.all(di_wet > di_dry)
        di_plus_50 = di + 50
        np.testing.assert_allclose(di_wet, di_plus_50, rtol=1e-03)
        di_minus_100 = di - 100
        np.testing.assert_allclose(di_dry, di_minus_100, rtol=1e-03)

        for value, array in {"200 mm": di, "250 mm": di_wet, "100 mm": di_dry}.items():
            assert array.attrs["long_name"] == "Growing season humidity"
            assert value in array.attrs["description"]


class TestPotentialEvapotranspiration:
    def test_convert_units(self, atmosds):
        ds = atmosds

        tn = ds.tasmin
        tx = ds.tasmax
        tm = ds.tas
        hurs = ds.hurs
        rsds = ds.rsds
        rsus = ds.rsus
        rlds = ds.rlds
        rlus = ds.rlus
        uas = ds.uas
        vas = ds.vas

        sfcWind, _ = atmos.wind_speed_from_vector(uas, vas)

        with xr.set_options(keep_attrs=True):
            tnC = tn - K2C
            tnC.attrs["units"] = "degC"
            tmC = tm - K2C
            tmC.attrs["units"] = "degC"

        pet_br65 = atmos.potential_evapotranspiration(tn, tx, method="BR65")
        pet_br65C = atmos.potential_evapotranspiration(tnC, tx, method="BR65")
        pet_hg85 = atmos.potential_evapotranspiration(tn, tx, method="HG85")
        pet_hg85C = atmos.potential_evapotranspiration(tnC, tx, method="HG85")
        pet_tw48 = atmos.potential_evapotranspiration(tas=tm, method="TW48")
        pet_tw48C = atmos.potential_evapotranspiration(tas=tmC, method="TW48")
        pet_mb05 = atmos.potential_evapotranspiration(tn, tx, method="MB05")
        pet_mb05C = atmos.potential_evapotranspiration(tnC, tx, method="MB05")
        pet_fao_pm98 = atmos.potential_evapotranspiration(
            tn,
            tx,
            hurs=hurs,
            rsds=rsds,
            rsus=rsus,
            rlds=rlds,
            rlus=rlus,
            sfcWind=sfcWind,
            method="FAO_PM98",
        )
        pet_fao_pm98C = atmos.potential_evapotranspiration(
            tnC,
            tx,
            hurs=hurs,
            rsds=rsds,
            rsus=rsus,
            rlds=rlds,
            rlus=rlus,
            sfcWind=sfcWind,
            method="FAO_PM98",
        )

        np.testing.assert_allclose(pet_br65, pet_br65C, atol=1)
        np.testing.assert_allclose(pet_hg85, pet_hg85C, atol=1)
        np.testing.assert_allclose(pet_tw48, pet_tw48C, atol=1)
        np.testing.assert_allclose(pet_mb05, pet_mb05C, atol=1)
        np.testing.assert_allclose(pet_fao_pm98, pet_fao_pm98C, atol=1)

    def test_nan_values(self, atmosds):
        ds = atmosds

        tn = ds.tasmin
        tx = ds.tasmax
        tm = ds.tas
        hurs = ds.hurs
        rsds = ds.rsds
        rsus = ds.rsus
        rlds = ds.rlds
        rlus = ds.rlus
        uas = ds.uas
        vas = ds.vas

        sfcWind, _ = atmos.wind_speed_from_vector(uas, vas)

        tn[0, 100] = np.nan
        tx[0, 101] = np.nan

        pet_br65 = atmos.potential_evapotranspiration(tn, tx, method="BR65")
        pet_hg85 = atmos.potential_evapotranspiration(tn, tx, method="HG85")
        pet_fao_pm98 = atmos.potential_evapotranspiration(
            tn,
            tx,
            hurs=hurs,
            rsds=rsds,
            rsus=rsus,
            rlds=rlds,
            rlus=rlus,
            sfcWind=sfcWind,
            method="FAO_PM98",
        )

        tm[0, 0:31] = np.nan

        pet_tw48 = atmos.potential_evapotranspiration(tas=tm, method="TW48")

        np.testing.assert_allclose(
            pet_br65.isel(location=0, time=slice(100, 102)), [np.nan, np.nan]
        )
        np.testing.assert_allclose(
            pet_hg85.isel(location=0, time=slice(100, 102)), [np.nan, np.nan]
        )
        np.testing.assert_allclose(
            pet_fao_pm98.isel(location=0, time=slice(100, 102)),
            [np.nan, np.nan],
        )
        np.testing.assert_allclose(pet_tw48.isel(location=0, time=0), [np.nan])


class TestWaterBudget:
    def test_convert_units(self, atmosds):
        ds = atmosds

        tn = ds.tasmin
        tx = ds.tasmax
        pr = ds.pr
        pet = ds.evspsblpot
        hurs = ds.hurs
        rsds = ds.rsds
        rsus = ds.rsus
        rlds = ds.rlds
        rlus = ds.rlus
        uas = ds.uas
        vas = ds.vas

        sfcWind, _ = atmos.wind_speed_from_vector(uas, vas)

        with xr.set_options(keep_attrs=True):
            tnC = tn - K2C
            tnC.attrs["units"] = "degC"
            prR = pr * 86400
            prR.attrs["units"] = "mm/day"
            petR = pet * 86400
            petR.attrs["units"] = "mm/day"

        p_pet_br65 = atmos.water_budget_from_tas(
            pr, tasmin=tn, tasmax=tx, method="BR65"
        )
        p_pet_br65C = atmos.water_budget_from_tas(
            prR, tasmin=tnC, tasmax=tx, method="BR65"
        )
        p_pet_hg85 = atmos.water_budget_from_tas(
            pr, tasmin=tn, tasmax=tx, method="HG85"
        )
        p_pet_hg85C = atmos.water_budget_from_tas(
            prR, tasmin=tnC, tasmax=tx, method="HG85"
        )
        p_pet_tw48 = atmos.water_budget_from_tas(
            pr, tasmin=tn, tasmax=tx, method="TW48"
        )
        p_pet_tw48C = atmos.water_budget_from_tas(
            prR, tasmin=tnC, tasmax=tx, method="TW48"
        )

        p_pet_fao_pm98 = atmos.water_budget_from_tas(
            pr=pr,
            tasmin=tn,
            tasmax=tx,
            hurs=hurs,
            rsds=rsds,
            rsus=rsus,
            rlds=rlds,
            rlus=rlus,
            sfcWind=sfcWind,
            method="FAO_PM98",
        )
        p_pet_fao_pm98R = atmos.water_budget_from_tas(
            pr=prR,
            tasmin=tn,
            tasmax=tx,
            hurs=hurs,
            rsds=rsds,
            rsus=rsus,
            rlds=rlds,
            rlus=rlus,
            sfcWind=sfcWind,
            method="FAO_PM98",
        )

        p_pet_evpot = atmos.water_budget(pr, evspsblpot=pet)
        p_pet_evpotR = atmos.water_budget(prR, evspsblpot=petR)

        np.testing.assert_allclose(p_pet_br65, p_pet_br65C, atol=1)
        np.testing.assert_allclose(p_pet_hg85, p_pet_hg85C, atol=1)
        np.testing.assert_allclose(p_pet_tw48, p_pet_tw48C, atol=1)
        np.testing.assert_allclose(p_pet_fao_pm98, p_pet_fao_pm98R, atol=1)
        np.testing.assert_allclose(p_pet_evpot, p_pet_evpotR, atol=1)

    def test_nan_values(self, atmosds):
        ds = atmosds

        tn = ds.tasmin
        tx = ds.tasmax
        tm = ds.tas
        pr = ds.pr
        pet = ds.evspsblpot
        hurs = ds.hurs
        rsds = ds.rsds
        rsus = ds.rsus
        rlds = ds.rlds
        rlus = ds.rlus
        uas = ds.uas
        vas = ds.vas

        sfcWind, _ = atmos.wind_speed_from_vector(uas, vas)

        tn[0, 100] = np.nan
        tx[0, 101] = np.nan

        p_pet_br65 = atmos.water_budget_from_tas(
            pr, tasmin=tn, tasmax=tx, method="BR65"
        )
        p_pet_hg85 = atmos.water_budget_from_tas(
            pr, tasmin=tn, tasmax=tx, method="HG85"
        )
        p_pet_fao_pm98 = atmos.water_budget_from_tas(
            pr=pr,
            tasmin=tn,
            tasmax=tx,
            hurs=hurs,
            rsds=rsds,
            rsus=rsus,
            rlds=rlds,
            rlus=rlus,
            sfcWind=sfcWind,
            method="FAO_PM98",
        )

        tm[0, 0:31] = np.nan

        p_pet_tw48 = atmos.water_budget_from_tas(pr, tas=tm, method="TW48")

        pet[0, 0:31] = np.nan

        p_pet_evpot = atmos.water_budget(pr, evspsblpot=pet)

        np.testing.assert_allclose(p_pet_br65[0, 100:102], [np.nan, np.nan])
        np.testing.assert_allclose(p_pet_hg85[0, 100:102], [np.nan, np.nan])
        np.testing.assert_allclose(p_pet_fao_pm98[0, 100:102], [np.nan, np.nan])
        np.testing.assert_allclose(p_pet_tw48[0, 0], [np.nan])
        np.testing.assert_allclose(p_pet_evpot[0, 0], [np.nan])


class TestUTCI:
    def test_universal_thermal_climate_index(self, atmosds):
        dataset = atmosds

        tas = dataset.tas
        hurs = dataset.hurs
        sfcWind, _sfcWindfromdir = atmos.wind_speed_from_vector(
            uas=dataset.uas, vas=dataset.vas
        )
        rsds = dataset.rsds
        rsus = dataset.rsus
        rlds = dataset.rlds
        rlus = dataset.rlus
        # Expected values
        utci_exp = [253.887, 254.367, 238.196, 252.826, 264.495]

        utci = atmos.universal_thermal_climate_index(
            tas=tas,
            hurs=hurs,
            sfcWind=sfcWind,
            rsds=rsds,
            rsus=rsus,
            rlds=rlds,
            rlus=rlus,
            stat="sunlit",
        )

        np.testing.assert_allclose(utci.isel(time=0), utci_exp, rtol=1e-03)


class TestMeanRadiantTemperature:
    def test_mean_radiant_temperature(self, atmosds):
        dataset = atmosds

        rsds = dataset.rsds
        rsus = dataset.rsus
        rlds = dataset.rlds
        rlus = dataset.rlus

        # Expected values
        exp_sun = [276.911, 274.742, 243.202, 268.012, 278.505]
        exp_ins = [276.911, 274.742, 243.202, 268.012, 278.505]

        mrt_sun = atmos.mean_radiant_temperature(rsds, rsus, rlds, rlus, stat="sunlit")
        mrt_ins = atmos.mean_radiant_temperature(rsds, rsus, rlds, rlus, stat="instant")

        rtol = 1e-03
        np.testing.assert_allclose(mrt_sun.isel(time=0), exp_sun, rtol=rtol)
        np.testing.assert_allclose(mrt_ins.isel(time=0), exp_ins, rtol=rtol)


class TestLateFrostDays:
    def test_late_frost_days(self, atmosds):
        tasmin = atmosds.tasmin

        # Expected values
        exp = [2, 9, 72, 24, 0]

        lfd = atmos.late_frost_days(tasmin, date_bounds=("04-01", "06-30"))

        np.testing.assert_allclose(lfd.isel(time=0), exp, rtol=1e-03)


def test_chill_units(atmosds):
    tasmax = atmosds.tasmax
    tasmin = atmosds.tasmin
    tas = make_hourly_temperature(tasmin, tasmax)
    cu = atmos.chill_units(tas, date_bounds=("04-01", "06-30"))
    assert cu.attrs["units"] == "1"
    assert cu.name == "cu"
    assert cu.time.size == 4

    # Values are confirmed with chillR package although not an exact match
    # due to implementation details
    exp = [1546.5, 1344.0, 1162.0, 1457.5]
    np.testing.assert_allclose(cu.isel(location=0), exp, rtol=1e-03)


@pytest.mark.parametrize("use_dask", [True, False])
def test_chill_portions(atmosds, use_dask):
    pytest.importorskip("flox")
    tasmax = atmosds.tasmax
    tasmin = atmosds.tasmin
    tas = make_hourly_temperature(tasmin, tasmax)
    if use_dask:
        tas = tas.chunk(time=tas.time.size // 2, location=1)

    with set_options(resample_map_blocks=True):
        cp = atmos.chill_portions(tas, date_bounds=("09-01", "03-30"), freq="YS-JUL")

    assert cp.attrs["units"] == "1"
    assert cp.name == "cp"
    # Although its 4 years of data its 5 seasons starting in July
    assert cp.time.size == 5

    # Values are confirmed with chillR package although not an exact match
    # due to implementation details
    exp = [np.nan, 99.91534493, 92.5473925, 99.03177047, np.nan]
    np.testing.assert_allclose(cp.isel(location=0), exp, rtol=1e-03)


def test_water_cycle_intensity(pr_series, evspsbl_series):
    pr = pr_series(np.ones(31))
    evspsbl = evspsbl_series(np.ones(31))

    wci = atmos.water_cycle_intensity(pr=pr, evspsbl=evspsbl, freq="MS")
    np.testing.assert_allclose(wci, 2 * 60 * 60 * 24 * 31)



from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

from xclim import atmos
from xclim.core.options import set_options
from xclim.core.units import convert_units_to
from xclim.indices.fire import (
    build_up_index,
    fire_season,
    fire_weather_index,
    fire_weather_ufunc,
    initial_spread_index,
    overwintering_drought_code,
)
from xclim.indices.fire._cffwis import (
    _day_length,
    _day_length_factor,
    _drought_code,
    _duff_moisture_code,
    _fine_fuel_moisture_code,
    _overwintering_drought_code,
)
from xclim.indices.run_length import run_bounds


class TestCFFWIS:
    # The following were computed with cffdrs 1.8.18, on the test_wDC data.
    cffdrs_fire_season = {
        "id1_default": [["2013-03-15", "2014-03-14"], ["2013-11-23", "2014-11-14"]],
        "id2_default": [["1980-04-20", "1981-05-15"], ["1980-10-16", "1981-10-14"]],
        "id3_default": [["1999-05-02", "2000-06-16"], ["1999-10-20", "2000-10-07"]],
        "id1_start10_end3": [
            ["2013-03-12", "2014-03-09", "2014-12-13"],
            ["2013-11-23", "2014-11-15", "2014-12-18"],
        ],
        "id1_start10_end3_YS": [
            ["2013-03-12", "2014-03-09"],
            ["2013-11-23", "2014-11-15"],
        ],
    }

    fwi_test_dataset = "FWI/cffdrs_test_fwi.nc"

    @classmethod
    def _get_cffdrs_fire_season(cls, key=None):
        def to_xr(arr):
            return xr.DataArray(
                np.array(arr, dtype=np.datetime64), dims=("bounds", "events")
            )

        if key:
            return to_xr(cls.cffdrs_fire_season[key])
        return {key: to_xr(arr) for key, arr in cls.cffdrs_fire_season.items()}

    def test_fine_fuel_moisture_code(self, open_dataset):
        fwi_data = open_dataset(self.fwi_test_dataset)
        ffmc = np.full(fwi_data.time.size + 1, np.nan)
        ffmc[0] = 85
        for i, t in enumerate(fwi_data.time):
            ffmc[i + 1] = _fine_fuel_moisture_code(
                fwi_data.sel(time=t).tas.values,
                fwi_data.sel(time=t).pr.values,
                fwi_data.sel(time=t).ws.values,
                fwi_data.sel(time=t).rh.values,
                ffmc[i],
            )

        np.testing.assert_allclose(ffmc[1:], fwi_data.ffmc.isel(test=0), rtol=1e-6)

    def test_duff_moisture_code(self, open_dataset):
        fwi_data = open_dataset(self.fwi_test_dataset)
        dmc = np.full(fwi_data.time.size + 1, np.nan)
        dmc[0] = 6
        for i, t in enumerate(fwi_data.time):
            dmc[i + 1] = _duff_moisture_code(
                fwi_data.sel(time=t).tas.values,
                fwi_data.sel(time=t).pr.values,
                fwi_data.sel(time=t).rh.values,
                fwi_data.sel(time=t).time.dt.month.values,
                fwi_data.lat.values,
                dmc[i],
            )

        np.testing.assert_allclose(dmc[1:], fwi_data.dmc.isel(test=0), rtol=1e-6)

    def test_drought_code(self, open_dataset):
        fwi_data = open_dataset(self.fwi_test_dataset)
        dc = np.full(fwi_data.time.size + 1, np.nan)
        dc[0] = 15
        for i, t in enumerate(fwi_data.time):
            dc[i + 1] = _drought_code(
                fwi_data.sel(time=t).tas.values,
                fwi_data.sel(time=t).pr.values,
                fwi_data.sel(time=t).time.dt.month.values,
                fwi_data.lat.values,
                dc[i],
            )

        np.testing.assert_allclose(dc[1:], fwi_data.dc.isel(test=0), rtol=1e-6)

    def test_initial_spread_index(self, open_dataset):
        fwi_data = open_dataset(self.fwi_test_dataset)
        isi = np.full(fwi_data.time.size, np.nan)
        for i, t in enumerate(fwi_data.time):
            isi[i] = initial_spread_index(
                fwi_data.sel(time=t).ws.values,
                fwi_data.sel(time=t).isel(test=0).ffmc.values,
            )
        np.testing.assert_allclose(isi, fwi_data.isi.isel(test=0), rtol=1e-6)

    def test_build_up_index(self, open_dataset):
        fwi_data = open_dataset(self.fwi_test_dataset)
        bui = np.full(fwi_data.time.size, np.nan)
        for i, t in enumerate(fwi_data.time):
            bui[i] = build_up_index(
                fwi_data.sel(time=t).isel(test=0).dmc.values,
                fwi_data.sel(time=t).isel(test=0).dc.values,
            )
        np.testing.assert_allclose(bui, fwi_data.bui.isel(test=0), rtol=1e-6)

    @pytest.mark.parametrize(
        "inputs,exp",
        [
            ([300, 110, 0.75, 0.75, 15], 109.4657),
            ([300, 110, 1.0, 0.9, 15], 16.35315),
            ([100, 50, 0.75, 0.75, 15], 105.176),
            ([1, 550, 0.75, 0.75, 10], 10),
        ],
    )
    def test_overwintering_drought_code(self, inputs, exp):
        wDC = _overwintering_drought_code(*inputs)
        np.testing.assert_allclose(wDC, exp, rtol=1e-6)

    @pytest.mark.parametrize(
        "inputs,exp",
        [
            ([300, 110, 0.75, 0.75, 15], 109.4657),
            ([300, 110, 1.0, 0.9, 15], 16.35315),
            ([100, 50, 0.75, 0.75, 15], 105.176),
            ([1, 550, 0.75, 0.75, 10], 10),
        ],
    )
    def test_overwintering_drought_code_indice(self, inputs, exp):
        last_dc = xr.DataArray([inputs[0]], dims=("x",), attrs={"units": ""})
        winter_pr = xr.DataArray([inputs[1]], dims=("x",), attrs={"units": "mm"})

        out = overwintering_drought_code(last_dc, winter_pr, *inputs[2:])

        np.testing.assert_allclose(out, exp, rtol=1e-6)

    def test_fire_weather_index(self, open_dataset):
        fwi_data = open_dataset(self.fwi_test_dataset)
        fwi = np.full(fwi_data.time.size, np.nan)
        for i, t in enumerate(fwi_data.time):
            fwi[i] = fire_weather_index(
                fwi_data.sel(time=t).isel(test=0).isi.values,
                fwi_data.sel(time=t).isel(test=0).bui.values,
            )
        np.testing.assert_allclose(fwi, fwi_data.fwi.isel(test=0), rtol=1e-6)

    def test_day_length(self):
        assert _day_length(44, 1) == 6.5

    def test_day_length_factor(self):
        assert _day_length_factor(44, 1) == -1.6

    def test_cffwis_indicator(self, open_dataset):
        fwi_data = open_dataset(self.fwi_test_dataset)
        fwi_data.lat.attrs["units"] = "degrees_north"
        dc, dmc, ffmc, isi, bui, fwi = atmos.cffwis_indices(
            tas=fwi_data.tas,
            pr=fwi_data.pr,
            hurs=fwi_data.rh,
            sfcWind=fwi_data.ws,
            lat=fwi_data.lat,
        )

        dc2, dmc2, ffmc2, isi2, bui2, fwi2 = atmos.cffwis_indices(
            tas=fwi_data.tas,
            pr=fwi_data.pr,
            hurs=fwi_data.rh,
            sfcWind=fwi_data.ws,
            lat=fwi_data.lat,
            ffmc0=ffmc[-1],
            dmc0=dmc[-1],
            dc0=dc[-1],
        )
        xr.testing.assert_allclose(dc, fwi_data.dc.isel(test=0), rtol=1e-6)
        xr.testing.assert_allclose(dmc, fwi_data.dmc.isel(test=0), rtol=1e-6)
        xr.testing.assert_allclose(ffmc, fwi_data.ffmc.isel(test=0), rtol=1e-6)
        xr.testing.assert_allclose(isi, fwi_data.isi.isel(test=0), rtol=1e-6)
        xr.testing.assert_allclose(bui, fwi_data.bui.isel(test=0), rtol=1e-6)
        xr.testing.assert_allclose(fwi, fwi_data.fwi.isel(test=0), rtol=1e-6)
        xr.testing.assert_allclose(dc2, fwi_data.dc.isel(test=1), rtol=1e-6)
        xr.testing.assert_allclose(dmc2, fwi_data.dmc.isel(test=1), rtol=1e-6)
        xr.testing.assert_allclose(ffmc2, fwi_data.ffmc.isel(test=1), rtol=1e-6)
        xr.testing.assert_allclose(isi2, fwi_data.isi.isel(test=1), rtol=1e-6)
        xr.testing.assert_allclose(bui2, fwi_data.bui.isel(test=1), rtol=1e-6)
        xr.testing.assert_allclose(fwi2, fwi_data.fwi.isel(test=1), rtol=1e-6)

    def test_fire_weather_ufunc_overwintering(self, atmosds):
        ds = atmosds.assign(
            tas=convert_units_to(atmosds.tas, "degC"),
            pr=convert_units_to(atmosds.pr, "mm/d", context="hydro"),
        )
        season_mask_all = fire_season(ds.tas, method="WF93", temp_end_thresh="4 degC")
        season_mask_all_LA08 = fire_season(ds.tas, snd=ds.swe, method="LA08")
        season_mask_yr = fire_season(ds.tas, method="WF93", freq="YS")

        # Mask is computed correctly and parameters are passed
        # season not passed so computed on the fly
        out1 = fire_weather_ufunc(
            tas=ds.tas,
            pr=ds.pr,
            lat=ds.lat,
            season_method="WF93",
            overwintering=False,
            temp_end_thresh=4,
            indexes=["DC"],
        )
        np.testing.assert_array_equal(out1["season_mask"], season_mask_all)

        out2 = fire_weather_ufunc(
            tas=ds.tas,
            pr=ds.pr,
            snd=ds.swe,
            lat=ds.lat,
            season_method="LA08",
            overwintering=True,
            indexes=["DC"],
        )
        np.testing.assert_array_equal(out2["season_mask"], season_mask_all_LA08)

        # Overwintering
        # Get last season's DC (from previous comp) and mask Saskatoon and Victoria
        dc0 = (
            out2["DC"]
            .ffill("time")
            .isel(time=-1)
            .where([True, True, True, False, False])
        )
        winter_pr = out2["winter_pr"]

        out3 = fire_weather_ufunc(
            tas=ds.tas,
            pr=ds.pr,
            lat=ds.lat,
            winter_pr=winter_pr,
            season_mask=season_mask_yr,
            dc0=dc0,
            overwintering=True,
            indexes=["DC"],
        )
        np.testing.assert_allclose(
            out3["winter_pr"].isel(location=0), 262.67575, rtol=1e-6
        )
        np.testing.assert_array_equal(out3["DC"].notnull(), season_mask_yr)

    def test_fire_weather_ufunc_drystart(self, atmosds):
        # This test is very shallow only tests if it runs.
        ds = atmosds.assign(
            tas=convert_units_to(atmosds.tas, "degC"),
            pr=convert_units_to(atmosds.pr, "mm/d", context="hydro"),
        )
        season_mask_yr = fire_season(ds.tas, method="WF93", freq="YS")

        out_ds = fire_weather_ufunc(
            tas=ds.tas,
            pr=ds.pr,
            hurs=ds.hurs,
            lat=ds.lat,
            season_mask=season_mask_yr,
            overwintering=False,
            dry_start="CFS",
            indexes=["DC", "DMC"],
            dmc_dry_factor=5,
        )
        out_no = fire_weather_ufunc(
            tas=ds.tas,
            pr=ds.pr,
            hurs=ds.hurs,
            lat=ds.lat,
            season_mask=season_mask_yr,
            overwintering=False,
            dry_start=None,
            indexes=["DC", "DMC"],
        )

        # I know season of 1992 is a "wet" start.
        xr.testing.assert_identical(
            out_ds["DC"].sel(location="Montréal", time="1992"),
            out_no["DC"].sel(location="Montréal", time="1992"),
        )
        xr.testing.assert_identical(
            out_ds["DMC"].sel(location="Montréal", time="1992"),
            out_no["DMC"].sel(location="Montréal", time="1992"),
        )

    def test_fire_weather_ufunc_errors(
        self, tas_series, pr_series, hurs_series, sfcWind_series
    ):
        tas = tas_series(np.ones(100), start="2017-01-01")
        pr = pr_series(np.ones(100), start="2017-01-01")
        hurs = hurs_series(np.ones(100), start="2017-01-01")
        sfcWind = sfcWind_series(np.ones(100), start="2017-01-01")

        snd = xr.full_like(tas, 0)
        lat = xr.full_like(tas.isel(time=0), 45)
        DC0 = xr.full_like(tas.isel(time=0), np.nan)  # noqa
        DMC0 = xr.full_like(tas.isel(time=0), np.nan)  # noqa
        FFMC0 = xr.full_like(tas.isel(time=0), np.nan)  # noqa

        # Test invalid combination
        with pytest.raises(TypeError):
            fire_weather_ufunc(
                tas=tas,
                pr=pr,
                hurs=hurs,
                lat=lat,
                dc0=DC0,
                indexes=["DC", "ISI"],
            )

        # Test missing arguments
        with pytest.raises(TypeError):
            fire_weather_ufunc(
                tas=tas,
                pr=pr,
                dc0=DC0,
                indexes=["DC"],  # lat=lat,
            )

        with pytest.raises(TypeError):
            fire_weather_ufunc(
                tas=tas,
                pr=pr,
                lat=lat,
                dc0=DC0,
                indexes=["DC"],
                season_method="LA08",
            )

        # Test output is complete + dask
        out = fire_weather_ufunc(
            tas=tas.chunk(),
            pr=pr.chunk(),
            lat=lat.chunk(),
            dc0=DC0,
            indexes=["DC"],
        )
        assert len(out.keys()) == 1
        out["DC"].load()

        out = fire_weather_ufunc(
            tas=tas,
            pr=pr,
            hurs=hurs,
            sfcWind=sfcWind,
            lat=lat,
            snd=snd,
            dc0=DC0,
            dmc0=DMC0,
            ffmc0=FFMC0,
            indexes=["DSR"],
        )

        assert len(out.keys()) == 7

    @pytest.mark.parametrize(
        "key,kwargs",
        [
            ("id1_default", {}),
            ("id2_default", {}),
            ("id3_default", {}),
            (
                "id1_start10_end3",
                {"temp_start_thresh": "283.15 K", "temp_end_thresh": "3 degC"},
            ),
            (
                "id1_start10_end3_YS",
                {
                    "temp_start_thresh": "283.15 K",
                    "temp_end_thresh": "3 degC",
                    "freq": "YS",
                },
            ),
        ],
    )
    def test_fire_season_R(self, open_dataset, key, kwargs):
        expected = self._get_cffdrs_fire_season(key)
        in_ds = open_dataset("FWI/cffdrs_test_wDC.nc")
        nid = int(key[2])

        mask = fire_season(tas=in_ds.where(in_ds.id == nid, drop=True).tasmax, **kwargs)
        bounds = run_bounds(mask, dim="time", coord=True)
        np.testing.assert_array_equal(bounds, expected)

    def test_gfwed_and_indicators(self, open_dataset):
        # Also tests passing parameters as quantity strings
        ds = open_dataset("FWI/GFWED_sample_2017.nc")

        outs = atmos.cffwis_indices(
            tas=ds.tas,
            pr=ds.prbc,
            snd=ds.snow_depth,
            hurs=ds.rh,
            sfcWind=ds.sfcwind,
            lat=ds.lat,
            season_method="GFWED",
            overwintering=False,
            dry_start="GFWED",
            temp_condition_days=3,
            snow_condition_days=3,
            temp_start_thresh="6 degC",
            temp_end_thresh="6 degC",
        )

        for exp, out in zip(
            [ds.DC, ds.DMC, ds.FFMC, ds.ISI, ds.BUI, ds.FWI], outs, strict=False
        ):
            np.testing.assert_allclose(
                out.isel(loc=[0, 1]), exp.isel(loc=[0, 1]), rtol=0.03
            )

        ds2 = ds.isel(time=slice(1, None))

        with set_options(cf_compliance="log"):
            mask = atmos.fire_season(
                tas=ds2.tas,
                snd=ds2.snow_depth,
                method="GFWED",
                temp_condition_days=3,
                snow_condition_days=3,
                temp_start_thresh="6 degC",
                temp_end_thresh="6 degC",
            )
            # 3 first days are false by default assume same as 4th day.
            mask = mask.where(mask.time > mask.time[2]).bfill("time")

            outs = atmos.cffwis_indices(
                tas=ds2.tas,
                pr=ds2.prbc,
                snd=ds2.snow_depth,
                hurs=ds2.rh,
                sfcWind=ds2.sfcwind,
                lat=ds2.lat,
                dc0=ds.DC.isel(time=0),
                dmc0=ds.DMC.isel(time=0),
                ffmc0=ds.FFMC.isel(time=0),
                season_mask=mask,
                overwintering=False,
                dry_start="GFWED",
                initial_start_up=False,
            )

        for exp, out in zip(
            [ds2.DC, ds2.DMC, ds2.FFMC, ds2.ISI, ds2.BUI, ds2.FWI], outs, strict=False
        ):
            np.testing.assert_allclose(out, exp, rtol=0.03)

    def test_gfwed_drought_code(self, open_dataset):
        # Also tests passing parameters as quantity strings
        ds = open_dataset("FWI/GFWED_sample_2017.nc")

        out = atmos.drought_code(
            tas=ds.tas,
            pr=ds.prbc,
            snd=ds.snow_depth,
            lat=ds.lat,
            season_method="GFWED",
            overwintering=False,
            dry_start="GFWED",
            temp_condition_days=3,
            snow_condition_days=3,
            temp_start_thresh="6 degC",
            temp_end_thresh="6 degC",
        )

        np.testing.assert_allclose(
            out.isel(loc=[0, 1]), ds.DC.isel(loc=[0, 1]), rtol=0.03
        )

    def test_gfwed_duff_moisture_code(self, open_dataset):
        # Also tests passing parameters as quantity strings
        ds = open_dataset("FWI/GFWED_sample_2017.nc")

        out = atmos.duff_moisture_code(
            tas=ds.tas,
            pr=ds.prbc,
            hurs=ds.rh,
            snd=ds.snow_depth,
            lat=ds.lat,
            season_method="GFWED",
            dry_start="GFWED",
            temp_condition_days=3,
            snow_condition_days=3,
            temp_start_thresh="6 degC",
            temp_end_thresh="6 degC",
        )

        np.testing.assert_allclose(
            out.isel(loc=[0, 1]), ds.DMC.isel(loc=[0, 1]), rtol=0.03
        )



"""Tests for generic indices."""

from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

from xclim.core.calendar import doy_to_days_since, select_time
from xclim.indices import generic
from xclim.testing.helpers import assert_lazy

K2C = 273.15


class TestSelectResampleOp:
    def test_month(self, q_series):
        q = q_series(np.arange(1000))
        o = generic.select_resample_op(q, "count", freq="YS", month=3)
        np.testing.assert_array_equal(o, 31)

    def test_season_default(self, q_series):
        # Will use freq='YS', so count J, F and D of each year.
        q = q_series(np.arange(1000))
        o = generic.select_resample_op(q, "min", season="DJF")
        assert o[0] == 0
        assert o[1] == 366

    def test_season(self, q_series):
        q = q_series(np.arange(1000))
        o = generic.select_resample_op(q, "count", freq="YS-DEC", season="DJF")
        assert o[0] == 31 + 29


class TestSelectRollingResampleOp:
    def test_rollingmax(self, q_series):
        q = q_series(np.arange(1, 366 + 365 + 365 + 1))  # 1st year is leap
        o = generic.select_rolling_resample_op(
            q, "max", window=14, window_center=False, window_op="mean"
        )
        np.testing.assert_array_equal(
            [
                np.mean(np.arange(353, 366 + 1)),
                np.mean(np.arange(353 + 365, 366 + 365 + 1)),
                np.mean(np.arange(353 + 365 * 2, 366 + 365 * 2 + 1)),
            ],
            o.values,
        )
        assert o.attrs["units"] == "m3 s-1"

    def test_rollingmaxindexer(self, q_series):
        q = q_series(np.arange(1, 366 + 365 + 365 + 1))  # 1st year is leap
        o = generic.select_rolling_resample_op(
            q, "min", window=14, window_center=False, window_op="max", season="DJF"
        )
        np.testing.assert_array_equal(
            [14, 367, 367 + 365], o.values
        )  # 14th day for 1st year, then Jan 1st for the next two
        assert o.attrs["units"] == "m3 s-1"

    def test_freq(self, q_series):
        q = q_series(np.arange(1, 366 + 365 + 365 + 1))  # 1st year is leap
        o = generic.select_rolling_resample_op(
            q, "max", window=3, window_center=True, window_op="integral", freq="MS"
        )
        np.testing.assert_array_equal(
            [
                np.sum([30, 31, 32]) * 86400,
                np.sum([30 + 29, 31 + 29, 32 + 29]) * 86400,
            ],  # m3 s-1 being summed by the frequency of the data
            o.isel(time=slice(0, 2)).values,
        )
        assert o.attrs["units"] == "m3"


class TestThresholdCount:
    def test_simple(self, tas_series):
        ts = tas_series(np.arange(365))
        out = generic.threshold_count(ts, "<", 50, "YE")
        np.testing.assert_array_equal(out, [50, 0])


class TestDomainCount:
    def test_simple(self, tas_series):
        ts = tas_series(np.arange(365))
        out = generic.domain_count(ts, low=10, high=20, freq="YE")
        np.testing.assert_array_equal(out, [10, 0])


class TestFlowGeneric:
    def test_doyminmax(self, q_series):
        a = np.ones(365)
        a[9] = 2
        a[19] = -2
        a[39] = 4
        a[49] = -4
        q = q_series(a)
        dmx = generic.doymax(q)
        dmn = generic.doymin(q)
        assert dmx.values == [40]
        assert dmn.values == [50]
        for da in [dmx, dmn]:
            for attr in ["units", "is_dayofyear", "calendar"]:
                assert attr in da.attrs.keys()

            assert da.attrs["units"] == "1"
            assert da.attrs["is_dayofyear"] == 1


class TestAggregateBetweenDates:
    def test_calendars(self):
        # generate test DataArray
        time_std = xr.date_range(
            "1991-07-01", "1993-06-30", freq="D", calendar="standard"
        )
        time_365 = xr.date_range(
            "1991-07-01", "1993-06-30", freq="D", calendar="noleap"
        )
        data_std = xr.DataArray(
            np.ones((time_std.size, 4)),
            dims=("time", "lon"),
            coords={"time": time_std, "lon": [-72, -71, -70, -69]},
        )
        # generate test start and end dates
        start_v = [[200, 200, np.nan, np.nan], [200, 200, 60, 60]]
        end_v = [[200, np.nan, 60, np.nan], [360, 60, 360, 80]]
        start_std = xr.DataArray(
            start_v,
            dims=("time", "lon"),
            coords={"time": [time_std[0], time_std[366]], "lon": data_std.lon},
            attrs={"calendar": "standard", "is_dayofyear": 1},
        )
        end_std = xr.DataArray(
            end_v,
            dims=("time", "lon"),
            coords={"time": [time_std[0], time_std[366]], "lon": data_std.lon},
            attrs={"calendar": "standard", "is_dayofyear": 1},
        )

        end_noleap = xr.DataArray(
            end_v,
            dims=("time", "lon"),
            coords={"time": [time_365[0], time_365[365]], "lon": data_std.lon},
            attrs={"calendar": "noleap", "is_dayofyear": 1},
        )

        out = generic.aggregate_between_dates(
            data_std, start_std, end_std, op="sum", freq="YS-JUL"
        )

        # expected output
        s = doy_to_days_since(start_std)
        e = doy_to_days_since(end_std)
        expected = e - s
        expected = xr.where(((s > e) | (s.isnull()) | (e.isnull())), np.nan, expected)

        np.testing.assert_allclose(out, expected)

        # check calendar conversion
        out_noleap = generic.aggregate_between_dates(
            data_std, start_std, end_noleap, op="sum", freq="YS-JUL"
        )

        np.testing.assert_allclose(out, out_noleap)

    def test_time_length(self):
        # generate test DataArray
        time_data = xr.date_range(
            "1991-01-01", "1993-12-31", freq="D", calendar="standard"
        )
        time_start = xr.date_range(
            "1990-01-01", "1992-12-31", freq="D", calendar="standard"
        )
        time_end = xr.date_range(
            "1991-01-01", "1993-12-31", freq="D", calendar="standard"
        )
        data = xr.DataArray(
            np.ones((time_data.size, 4)),
            dims=("time", "lon"),
            coords={"time": time_data, "lon": [-72, -71, -70, -69]},
        )
        # generate test start and end dates
        start_v = [[200, 200, np.nan, np.nan], [200, 200, 60, 60], [150, 100, 40, 10]]
        end_v = [[200, np.nan, 60, np.nan], [360, 60, 360, 80], [200, 200, 60, 50]]
        start = xr.DataArray(
            start_v,
            dims=("time", "lon"),
            coords={
                "time": [time_start[0], time_start[365], time_start[730]],
                "lon": data.lon,
            },
            attrs={"calendar": "standard", "is_dayofyear": 1},
        )
        end = xr.DataArray(
            end_v,
            dims=("time", "lon"),
            coords={
                "time": [time_end[0], time_end[365], time_end[731]],
                "lon": data.lon,
            },
            attrs={"calendar": "standard", "is_dayofyear": 1},
        )

        out = generic.aggregate_between_dates(data, start, end, op="sum", freq="YS")

        # expected output
        s = doy_to_days_since(start)
        e = doy_to_days_since(end)
        expected = e - s
        expected[1, 1] = np.nan

        np.testing.assert_allclose(out[0:2], expected)
        np.testing.assert_allclose(out[2], np.array([np.nan, np.nan, np.nan, np.nan]))

    def test_frequency(self):
        # generate test DataArray
        time_data = xr.date_range(
            "1991-01-01", "1992-05-31", freq="D", calendar="standard"
        )
        data = xr.DataArray(
            np.ones((time_data.size, 2)),
            dims=("time", "lon"),
            coords={"time": time_data, "lon": [-70, -69]},
        )
        # generate test start and end dates
        start_v = [[70, 100], [200, 200], [270, 300], [35, 35], [80, 80]]
        end_v = [[130, 70], [200, np.nan], [330, 270], [35, np.nan], [150, 150]]
        end_m_v = [[20, 20], [40, 40], [80, 80], [100, 100], [130, 130]]
        start = xr.DataArray(
            start_v,
            dims=("time", "lon"),
            coords={
                "time": [
                    time_data[59],
                    time_data[151],
                    time_data[243],
                    time_data[334],
                    time_data[425],
                ],
                "lon": data.lon,
            },
            attrs={"calendar": "standard", "is_dayofyear": 1},
        )
        end = xr.DataArray(
            end_v,
            dims=("time", "lon"),
            coords={
                "time": [
                    time_data[59],
                    time_data[151],
                    time_data[243],
                    time_data[334],
                    time_data[425],
                ],
                "lon": data.lon,
            },
            attrs={"calendar": "standard", "is_dayofyear": 1},
        )
        end_m = xr.DataArray(
            end_m_v,
            dims=("time", "lon"),
            coords={
                "time": [
                    time_data[0],
                    time_data[31],
                    time_data[59],
                    time_data[90],
                    time_data[120],
                ],
                "lon": data.lon,
            },
            attrs={"calendar": "standard", "is_dayofyear": 1},
        )

        out = generic.aggregate_between_dates(data, start, end, op="sum", freq="QS-DEC")

        # expected output
        s = doy_to_days_since(start)
        e = doy_to_days_since(end)
        expected = e - s
        expected = xr.where(expected < 0, np.nan, expected)

        np.testing.assert_allclose(out[0], np.array([np.nan, np.nan]))
        np.testing.assert_allclose(out[1:6], expected)

        with pytest.raises(ValueError):
            generic.aggregate_between_dates(data, start, end_m)

    def test_day_of_year_strings(self):
        # generate test DataArray
        time_data = xr.date_range(
            "1990-08-01", "1995-06-01", freq="D", calendar="standard"
        )
        data = xr.DataArray(
            np.ones(time_data.size),
            dims="time",
            coords={"time": time_data},
        )
        # set start and end dates
        start = "02-01"
        end = "10-31"

        out = generic.aggregate_between_dates(data, start, end, op="sum", freq="YS")

        np.testing.assert_allclose(out, np.array([np.nan, 272, 273, 272, 272, np.nan]))

        # given no freq and only strings for start and end dates
        with pytest.raises(ValueError):
            generic.aggregate_between_dates(data, start, end, op="sum")

        # given a malformed date string
        bad_start = "02-31"
        with pytest.raises(ValueError):
            generic.aggregate_between_dates(data, bad_start, end, op="sum", freq="YS")


class TestCumulativeDifference:
    @pytest.mark.parametrize(
        "op, expected",
        [("gt", [0, 5, 10, 0, 0]), (">=", [0, 5, 10, 0, 0]), ("<", [20, 0, 0, 7, 0])],
    )
    def test_simple(self, tas_series, op, expected):
        tas = tas_series(np.array([-10, 15, 20, 3, 10]) + K2C)

        out = generic.cumulative_difference(tas, threshold="10 degC", op=op)
        out_kelvin = generic.cumulative_difference(tas, threshold="283.15 degK", op=op)

        np.testing.assert_allclose(out, expected)
        np.testing.assert_allclose(out, out_kelvin)

    def test_forbidden(self, tas_series):
        tas = tas_series(np.array([-10, 15, 20, 3, 10]) + K2C)

        with pytest.raises(NotImplementedError):
            generic.cumulative_difference(tas, threshold="10 degC", op="!=")

    def test_delta_units(self, tas_series):
        tas = tas_series(np.array([-10, 15, 20, 3, 10]) + K2C)

        out = generic.cumulative_difference(tas, threshold="10 degC", op=">=")
        assert "units_metadata" in out.attrs


class TestFirstDayThreshold:
    @pytest.mark.parametrize(
        "op, expected",
        [(">", 6), (">=", 5), ("==", 5), ("!=", 1)],
    )
    def test_generic_precip_above(self, pr_series, op, expected):
        a = np.zeros(365)
        a[:8] = np.arange(8) / 1000
        pr = pr_series(a, start="1/1/2000")

        fda = generic.first_day_threshold_reached(
            pr,
            threshold="0.004 kg m-2 s-1",
            op=op,
            after_date="01-01",
            window=1,
            freq="YS",
        )
        assert fda == expected

    @pytest.mark.parametrize(
        "op, expected",
        [("lt", 5), ("le", 4), ("eq", 4), ("ne", 1)],
    )
    def test_generic_precip_below(self, pr_series, op, expected):
        a = np.zeros(365)
        precip = np.arange(8) / 1000
        a[:8] = np.flip(precip)
        pr = pr_series(a, start="1/1/2000")

        fdb = generic.first_day_threshold_reached(
            pr,
            threshold="0.004 kg m-2 s-1",
            op=op,
            after_date="01-01",
            window=1,
            freq="YS",
        )
        assert fdb == expected

    def test_generic_forbidden_op(self, pr_series):
        a = np.zeros(365)
        precip = np.arange(8) / 1000
        a[:8] = np.flip(precip)
        pr = pr_series(a, start="1/1/2000")

        with pytest.raises(ValueError):
            generic.first_day_threshold_reached(
                pr,
                threshold="0.004 kg m-2 s-1",
                op=">",
                after_date="01-01",
                window=1,
                freq="YS",
                constrain=("<", "<="),
            )


class TestGetDailyEvents:
    def test_simple(self, tas_series):
        arr = xr.DataArray(np.array([-10, 15, 20, np.nan, 10]), name="Stuff")

        out = generic.get_daily_events(arr, threshold=10, op=">=")

        assert out.name == "events"
        assert out.sum() == 3
        np.testing.assert_array_equal(out, [0, 1, 1, np.nan, 1])


class TestGenericCountingIndices:
    @pytest.mark.parametrize(
        "op_high, op_low, expected",
        [(">", "<", 1), (">", "<=", 2), (">=", "<", 3), (">=", "<=", 4)],
    )
    def test_simple_count_level_crossings(
        self, tasmin_series, tasmax_series, op_high, op_low, expected
    ):
        tasmin = tasmin_series(np.array([-1, -3, 0, 5, 9, 1, 3]) + K2C)
        tasmax = tasmax_series(np.array([5, 7, 3, 6, 13, 5, 4]) + K2C)

        crossings = generic.count_level_crossings(
            tasmin,
            tasmax,
            threshold="5 degC",
            freq="YS",
            op_high=op_high,
            op_low=op_low,
        )
        np.testing.assert_array_equal(crossings, [expected])

    @pytest.mark.parametrize(
        "op_high, op_low", [("<=", "<="), (">=", ">="), ("<", ">"), ("==", "!=")]
    )
    def test_forbidden_op(self, tasmin_series, tasmax_series, op_high, op_low):
        tasmin = tasmin_series(np.zeros(7) + K2C)
        tasmax = tasmax_series(np.ones(7) + K2C)

        with pytest.raises(ValueError):
            generic.count_level_crossings(
                tasmin,
                tasmax,
                threshold="0.5 degC",
                freq="YS",
                op_high=op_high,
                op_low=op_low,
            )

    @pytest.mark.parametrize(
        "op, constrain, expected, should_fail",
        [
            ("<", ("!=", "<"), 4, False),
            (">", (">", "<="), 5, False),
            (">=", (">=", "=="), 6, False),
            ("==", ("==", "!="), 1, False),
            ("==", (">", ">="), 1, True),
            ("!=", ("!=", ">"), 9, False),
            ("!=", (">", "=="), 9, True),
            ("%", ("%", "$", "@"), 5.29e-11, True),
        ],
    )
    def test_count_occurrences(self, tas_series, op, constrain, expected, should_fail):
        tas = tas_series(np.arange(10) + K2C)

        if should_fail:
            with pytest.raises(ValueError):
                generic.count_occurrences(
                    tas, "4 degC", freq="YS", op=op, constrain=constrain
                )
        else:
            occurrences = generic.count_occurrences(
                tas, "4 degC", freq="YS", op=op, constrain=constrain
            )
            np.testing.assert_array_equal(occurrences, [expected])

    @pytest.mark.parametrize(
        "op, constrain, expected, should_fail",
        [
            ("<", None, np.nan, False),
            ("<=", None, 3, False),
            ("!=", ("!=",), 1, False),
            ("==", ("==", "!="), 3, False),
            ("==", (">=", ">", "<"), 3, True),
        ],
    )
    def test_first_occurrence(self, tas_series, op, constrain, expected, should_fail):
        tas = tas_series(
            np.array([15, 12, 11, 12, 14, 13, 18, 11, 13]) + K2C, start="1/1/2000"
        )

        if should_fail:
            with pytest.raises(ValueError):
                generic.first_occurrence(
                    tas, threshold="11 degC", freq="YS", op=op, constrain=constrain
                )
        else:
            first = generic.first_occurrence(
                tas, threshold="11 degC", freq="YS", op=op, constrain=constrain
            )

            np.testing.assert_array_equal(first, [expected])

    @pytest.mark.parametrize(
        "op, constrain, expected, should_fail",
        [
            ("<", None, np.nan, False),
            ("<=", None, 8, False),
            ("!=", ("!=",), 9, False),
            ("==", ("==", "!="), 8, False),
            ("==", (">=", ">", "<"), 5, True),
        ],
    )
    def test_last_occurrence(self, tas_series, op, constrain, expected, should_fail):
        tas = tas_series(
            np.array([15, 12, 11, 12, 14, 13, 18, 11, 13]) + K2C, start="1/1/2000"
        )

        if should_fail:
            with pytest.raises(ValueError):
                generic.last_occurrence(
                    tas, threshold="11 degC", freq="YS", op=op, constrain=constrain
                )
        else:
            first = generic.last_occurrence(
                tas, threshold="11 degC", freq="YS", op=op, constrain=constrain
            )

            np.testing.assert_array_equal(first, [expected])


class TestTimeSelection:
    @staticmethod
    def series(start, end, calendar):
        time = xr.date_range(
            start,
            end,
            calendar=calendar.replace("default", "proleptic_gregorian"),
            use_cftime=(calendar != "default"),
        )
        return xr.DataArray([1] * time.size, dims=("time",), coords={"time": time})

    def test_select_time_month(self):
        da = self.series("1993-01-05", "1994-12-31", "default")

        out = select_time(da, drop=True, month=1)
        exp = xr.concat(
            (
                self.series("1993-01-05", "1993-01-31", "default"),
                self.series("1994-01-01", "1994-01-31", "default"),
            ),
            "time",
        )
        xr.testing.assert_equal(out, exp)

        out = select_time(da, month=1)
        xr.testing.assert_equal(out.time, da.time)
        assert out.sum() == 58

        da = self.series("1993-01-05", "1994-12-30", "360_day")
        out = select_time(da, drop=True, month=[3, 6])
        exp = xr.concat(
            (
                self.series("1993-03-01", "1993-03-30", "360_day"),
                self.series("1993-06-01", "1993-06-30", "360_day"),
                self.series("1994-03-01", "1994-03-30", "360_day"),
                self.series("1994-06-01", "1994-06-30", "360_day"),
            ),
            "time",
        )
        xr.testing.assert_equal(out, exp)

    def test_select_time_season(self):
        da = self.series("1993-01-05", "1994-12-31", "default")

        out = select_time(da, drop=True, season="DJF")
        exp = xr.concat(
            (
                self.series("1993-01-05", "1993-02-28", "default"),
                self.series("1993-12-01", "1994-02-28", "default"),
                self.series("1994-12-01", "1994-12-31", "default"),
            ),
            "time",
        )
        xr.testing.assert_equal(out, exp)

        da = self.series("1993-01-05", "1994-12-31", "365_day")
        out = select_time(da, drop=True, season=["MAM", "SON"])
        exp = xr.concat(
            (
                self.series("1993-03-01", "1993-05-31", "365_day"),
                self.series("1993-09-01", "1993-11-30", "365_day"),
                self.series("1994-03-01", "1994-05-31", "365_day"),
                self.series("1994-09-01", "1994-11-30", "365_day"),
            ),
            "time",
        )
        xr.testing.assert_equal(out, exp)

    def test_select_time_doys(self):
        da = self.series("2003-02-13", "2004-12-31", "default")

        out = select_time(da, drop=True, doy_bounds=(360, 75))
        exp = xr.concat(
            (
                self.series("2003-02-13", "2003-03-16", "default"),
                self.series("2003-12-26", "2004-03-15", "default"),
                self.series("2004-12-25", "2004-12-31", "default"),
            ),
            "time",
        )
        xr.testing.assert_equal(out, exp)

        da = self.series("2003-02-13", "2004-12-31", "proleptic_gregorian")

        out = select_time(da, drop=True, doy_bounds=(25, 80))
        exp = xr.concat(
            (
                self.series("2003-02-13", "2003-03-21", "proleptic_gregorian"),
                self.series("2004-01-25", "2004-03-20", "proleptic_gregorian"),
            ),
            "time",
        )
        xr.testing.assert_equal(out, exp)

    def test_select_time_dates(self):
        da = self.series("2003-02-13", "2004-11-01", "all_leap")
        da = da.where(da.time.dt.dayofyear != 92, drop=True)  # no 04-01

        out = select_time(da, drop=True, date_bounds=("04-01", "12-04"))
        exp = xr.concat(
            (
                self.series("2003-04-02", "2003-12-04", "all_leap"),
                self.series("2004-04-02", "2004-11-01", "all_leap"),
            ),
            "time",
        )
        xr.testing.assert_equal(out, exp)

        da = self.series("2003-02-13", "2005-11-01", "standard")

        out = select_time(da, drop=True, date_bounds=("10-05", "02-29"))
        exp = xr.concat(
            (
                self.series("2003-02-13", "2003-02-28", "standard"),
                self.series("2003-10-05", "2004-02-29", "standard"),
                self.series("2004-10-05", "2005-02-28", "standard"),
                self.series("2005-10-05", "2005-11-01", "standard"),
            ),
            "time",
        )
        xr.testing.assert_equal(out, exp)

    def test_select_time_errors(self):
        da = self.series("2003-01-01", "2004-01-01", "standard")

        xr.testing.assert_identical(da, select_time(da))

        with pytest.raises(
            ValueError, match="Only one method of indexing may be given"
        ):
            select_time(da, season="DJF", month=[3, 4, 5])

        with pytest.raises(ValueError, match="invalid day number provided in cftime."):
            select_time(da, date_bounds=("02-30", "03-03"))

        with pytest.raises(ValueError):
            select_time(da, date_bounds=("02-30",))

        with pytest.raises(TypeError):
            select_time(da, doy_bounds=(300, 203, 202))


class TestSpellMask:
    def test_single_variable(self):
        data = xr.DataArray([0, 1, 2, 3, 2, 1, 0, 0], dims=("time",))

        out = generic.spell_mask(data, 3, "min", ">=", 2)
        np.testing.assert_array_equal(
            out, np.array([0, 0, 1, 1, 1, 0, 0, 0]).astype(bool)
        )

        out = generic.spell_mask(data, 3, "max", ">=", 2)
        np.testing.assert_array_equal(
            out, np.array([1, 1, 1, 1, 1, 1, 1, 0]).astype(bool)
        )

        out = generic.spell_mask(data, 2, "mean", ">=", 2)
        np.testing.assert_array_equal(
            out, np.array([0, 0, 1, 1, 1, 0, 0, 0]).astype(bool)
        )

        out = generic.spell_mask(data, 3, "mean", ">", 2, weights=[0.2, 0.4, 0.4])
        np.testing.assert_array_equal(
            out, np.array([0, 1, 1, 1, 1, 0, 0, 0]).astype(bool)
        )

    def test_multiple_variables(self):
        data1 = xr.DataArray([0, 1, 2, 3, 2, 1, 0, 0], dims=("time",))
        data2 = xr.DataArray([1, 2, 3, 2, 1, 0, 0, 0], dims=("time",))

        out = generic.spell_mask([data1, data2], 3, "min", ">=", [2, 2])
        np.testing.assert_array_equal(
            out, np.array([0, 0, 0, 0, 0, 0, 0, 0]).astype(bool)
        )

        out = generic.spell_mask(
            [data1, data2], 3, "min", ">=", [2, 2], var_reducer="any"
        )
        np.testing.assert_array_equal(
            out, np.array([0, 1, 1, 1, 1, 0, 0, 0]).astype(bool)
        )

        out = generic.spell_mask([data1, data2], 2, "mean", ">=", [2, 2])
        np.testing.assert_array_equal(
            out, np.array([0, 0, 1, 1, 0, 0, 0, 0]).astype(bool)
        )

        out = generic.spell_mask(
            [data1, data2], 3, "mean", ">", [2, 1.5], weights=[0.2, 0.4, 0.4]
        )
        np.testing.assert_array_equal(
            out, np.array([0, 1, 1, 1, 1, 0, 0, 0]).astype(bool)
        )

    def test_errors(self):
        data = xr.DataArray([0, 1, 2, 3, 2, 1, 0, 0], dims=("time",))

        # Threshold must be seq
        with pytest.raises(ValueError, match="must be a sequence of the same length"):
            generic.spell_mask([data, data], 3, "min", "<=", 2)

        # Threshold must be same length
        with pytest.raises(ValueError, match="must be a sequence of the same length"):
            generic.spell_mask([data, data], 3, "min", "<=", [2])

        # Weights must have win_reducer = 'mean'
        with pytest.raises(
            ValueError, match="is only supported if 'win_reducer' is 'mean'"
        ):
            generic.spell_mask(data, 3, "min", "<=", 2, weights=[1, 2, 3])

        # Weights must have same length as window
        with pytest.raises(ValueError, match="Weights have a different length"):
            generic.spell_mask(data, 3, "mean", "<=", 2, weights=[1, 2])


def test_spell_length_statistics_multi(tasmin_series, tasmax_series):
    tn = tasmin_series(
        np.zeros(
            365,
        )
        + 270,
        start="2001-01-01",
    )
    tx = tasmax_series(
        np.zeros(
            365,
        )
        + 270,
        start="2001-01-01",
    )

    outc, outs, outm = generic.bivariate_spell_length_statistics(
        tn,
        "0 °C",
        tx,
        "1°C",
        window=5,
        win_reducer="min",
        op="<",
        spell_reducer=["count", "sum", "max"],
        freq="YS",
    )
    xr.testing.assert_equal(outs, outm)
    np.testing.assert_allclose(outc, 1)


class TestThresholdedEvents:

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_simple(self, pr_series, use_dask):
        arr = np.array([0, 0, 0, 1, 2, 3, 0, 3, 3, 10, 0, 0, 0, 0, 0, 1, 2, 2, 2, 0, 0, 0, 0, 0, 0, 1, 3, 3, 2, 0, 0, 0, 2, 0, 0, 0, 0])  # fmt: skip
        pr = pr_series(arr, start="2000-01-01", units="mm")
        if use_dask:
            pr = pr.chunk(-1)

        with assert_lazy:
            out = generic.thresholded_events(
                pr,
                thresh="1 mm",
                op=">=",
                window=3,
            )

        assert out.event.size == np.ceil(arr.size / (3 + 1))
        out = out.load().dropna("event", how="all")

        np.testing.assert_array_equal(out.event_length, [3, 3, 4, 4])
        np.testing.assert_array_equal(out.event_effective_length, [3, 3, 4, 4])
        np.testing.assert_array_equal(out.event_sum, [6, 16, 7, 9])
        np.testing.assert_array_equal(
            out.event_start,
            np.array(
                ["2000-01-04", "2000-01-08", "2000-01-16", "2000-01-26"],
                dtype="datetime64[ns]",
            ),
        )

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_diff_windows(self, pr_series, use_dask):
        arr = np.array([0, 0, 0, 1, 2, 3, 0, 3, 3, 10, 0, 0, 0, 0, 0, 1, 2, 2, 2, 0, 0, 0, 0, 0, 0, 1, 3, 3, 2, 0, 0, 0, 2, 0, 0, 0, 0])  # fmt: skip
        pr = pr_series(arr, start="2000-01-01", units="mm")
        if use_dask:
            pr = pr.chunk(-1)

        # different window stop
        out = (
            generic.thresholded_events(
                pr, thresh="2 mm", op=">=", window=3, window_stop=4
            )
            .load()
            .dropna("event", how="all")
        )

        np.testing.assert_array_equal(out.event_length, [3, 3, 7])
        np.testing.assert_array_equal(out.event_effective_length, [3, 3, 4])
        np.testing.assert_array_equal(out.event_sum, [16, 6, 10])
        np.testing.assert_array_equal(
            out.event_start,
            np.array(
                ["2000-01-08", "2000-01-17", "2000-01-27"], dtype="datetime64[ns]"
            ),
        )

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_cftime(self, pr_series, use_dask):
        arr = np.array([0, 0, 0, 1, 2, 3, 0, 3, 3, 10, 0, 0, 0, 0, 0, 1, 2, 2, 2, 0, 0, 0, 0, 0, 0, 1, 3, 3, 2, 0, 0, 0, 2, 0, 0, 0, 0])  # fmt: skip
        pr = pr_series(arr, start="2000-01-01", units="mm").convert_calendar("noleap")
        if use_dask:
            pr = pr.chunk(-1)

        # cftime
        with assert_lazy:
            out = generic.thresholded_events(
                pr,
                thresh="1 mm",
                op=">=",
                window=3,
                window_stop=3,
            )
        out = out.load().dropna("event", how="all")

        np.testing.assert_array_equal(out.event_length, [7, 4, 4])
        np.testing.assert_array_equal(out.event_effective_length, [6, 4, 4])
        np.testing.assert_array_equal(out.event_sum, [22, 7, 9])
        exp = xr.DataArray(
            [1, 2, 3],
            dims=("time",),
            coords={
                "time": np.array(
                    ["2000-01-04", "2000-01-16", "2000-01-26"], dtype="datetime64[ns]"
                )
            },
        )
        np.testing.assert_array_equal(
            out.event_start, exp.convert_calendar("noleap").time
        )

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_freq(self, pr_series, use_dask):
        jan = [0, 0, 0, 1, 2, 3, 0, 3, 3, 10, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 3, 2, 3, 2]  # fmt: skip
        fev = [2, 2, 1, 0, 0, 0, 3, 3, 4, 5, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]  # fmt: skip
        pr = pr_series(np.array(jan + fev), start="2000-01-01", units="mm")
        if use_dask:
            pr = pr.chunk(-1)

        with assert_lazy:
            out = generic.thresholded_events(
                pr, thresh="1 mm", op=">=", window=3, freq="MS", window_stop=3
            )
        assert out.event_length.shape == (2, 6)
        out = out.load().dropna("event", how="all")

        np.testing.assert_array_equal(out.event_length, [[7, 6, 4], [3, 5, np.nan]])
        np.testing.assert_array_equal(
            out.event_effective_length, [[6, 6, 4], [3, 5, np.nan]]
        )
        np.testing.assert_array_equal(out.event_sum, [[22, 12, 10], [5, 17, np.nan]])
        np.testing.assert_array_equal(
            out.event_start,
            np.array(
                [
                    ["2000-01-04", "2000-01-17", "2000-01-28"],
                    ["2000-02-01", "2000-02-07", "NaT"],
                ],
                dtype="datetime64[ns]",
            ),
        )



# noqa: D100
from __future__ import annotations

import xarray as xr

from xclim.core.units import declare_units, rate2amount


@declare_units(pr="[precipitation]")
def extreme_precip_accumulation_and_days(
    pr: xr.DataArray, perc: float = 95, freq: str = "YS"
) -> tuple[xr.DataArray, xr.DataArray]:
    """Total precipitation accumulation during extreme events and number of days of such precipitation.

    The `perc` percentile of the precipitation (including all values, not in a day-of-year manner)
    is computed. Then, for each period, the days where `pr` is above the threshold are accumulated,
    to get the total precip related to those extreme events.

    Parameters
    ----------
    pr: xr.DataArray
        Precipitation flux (both phases).
    perc: float
        Percentile corresponding to "extreme" precipitation, [0-100].
    freq: str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray
        Precipitation accumulated during events where pr was above the {perc}th percentile of the whole series.
    xarray.DataArray
        Number of days where pr was above the {perc}th percentile of the whole series.
    """
    pr_thresh = pr.quantile(perc / 100, dim="time").drop_vars("quantile")

    extreme_days = pr >= pr_thresh
    pr_extreme = rate2amount(pr).where(extreme_days)

    out1 = pr_extreme.resample(time=freq).sum()
    out1 = out1.assign_attrs(units=pr_extreme.units)

    out2 = extreme_days.resample(time=freq).sum()
    out2 = out2.assign_attrs(units="days")
    return out1, out2



"""
==============
SDBA submodule
==============
"""

from __future__ import annotations

from . import adjustment, detrending, measures, processing, properties, utils
from .adjustment import *
from .base import Grouper
from .processing import stack_variables, unstack_variables

# TODO: ISIMIP ? Used for precip freq adjustment in biasCorrection.R
# Hempel, S., Frieler, K., Warszawski, L., Schewe, J., & Piontek, F. (2013). A trend-preserving bias correction &ndash;
# The ISI-MIP approach. Earth System Dynamics, 4(2), 219–236. https://doi.org/10.5194/esd-4-219-2013
# If SBCK is installed, create adjustment classes wrapping SBCK's algorithms.
if hasattr(adjustment, "_generate_SBCK_classes"):
    for cls in adjustment._generate_SBCK_classes():
        adjustment.__dict__[cls.__name__] = cls



from __future__ import annotations

import logging
import warnings

logger = logging.getLogger("xclim")

__all__ = ["MissingVariableError", "ValidationError", "raise_warn_or_log"]


class ValidationError(ValueError):
    """Error raised when input data to an indicator fails the validation tests."""

    @property
    def msg(self):  # noqa
        return self.args[0]


class MissingVariableError(ValueError):
    """Error raised when a dataset is passed to an indicator but one of the needed variable is missing."""


def raise_warn_or_log(
    err: Exception,
    mode: str,
    msg: str | None = None,
    err_type: type = ValueError,
    stacklevel: int = 1,
):
    """Raise, warn or log an error according.

    Parameters
    ----------
    err : Exception
        An error.
    mode : {'ignore', 'log', 'warn', 'raise'}
        What to do with the error.
    msg : str, optional
        The string used when logging or warning.
        Defaults to the `msg` attr of the error (if present) or to "Failed with <err>".
    err_type : type
        The type of error/exception to raise.
    stacklevel : int
        Stacklevel when warning. Relative to the call of this function (1 is added).
    """
    message = msg or getattr(err, "msg", f"Failed with {err!r}.")
    if mode == "ignore":
        pass
    elif mode == "log":
        logger.info(message)
    elif mode == "warn":
        warnings.warn(message, stacklevel=stacklevel + 1)
    else:  # mode == "raise"
        raise err from err_type(message)



#!/usr/bin/env python
# Tests for `xclim.core.options`
from __future__ import annotations

import pytest

from xclim import set_options
from xclim.core.missing import MissingBase
from xclim.core.options import OPTIONS, register_missing_method


@pytest.mark.parametrize(
    "option,value",
    [
        ("metadata_locales", ["fr"]),
        ("data_validation", "log"),
        ("data_validation", "raise"),
        ("cf_compliance", "log"),
        ("cf_compliance", "raise"),
        ("check_missing", "wmo"),
        ("check_missing", "any"),
        ("missing_options", {"wmo": {"nm": 10, "nc": 3}}),
        ("missing_options", {"pct": {"tolerance": 0.1}}),
        ("missing_options", {"wmo": {"nm": 10, "nc": 3}, "pct": {"tolerance": 0.1}}),
    ],
)
def test_set_options_valid(option, value):
    old = OPTIONS[option]
    with set_options(**{option: value}):
        if option != "missing_options":
            assert OPTIONS[option] == value
        else:
            for k, opts in value.items():
                curr_opts = OPTIONS["missing_options"][k].copy()
                assert curr_opts == opts
    assert OPTIONS[option] == old


@pytest.mark.parametrize(
    "option,value",
    [
        ("metadata_locales", ["tlh"]),
        ("metadata_locales", [("tlh", "not/a/real/klingo/file.json")]),
        ("data_validation", True),
        ("check_missing", "from_context"),
        ("cf_compliance", False),
        ("missing_options", {"pct": {"nm": 45}}),
        ("missing_options", {"wmo": {"nm": 45, "nc": 3}}),
        (
            "missing_options",
            {"wmo": {"nm": 45, "nc": 3}, "notachoice": {"tolerance": 45}},
        ),
        (
            "missing_options",
            {"wmo": {"nm": 45, "nc": 3, "_validator": lambda x: x < 1}},
        ),
    ],
)
def test_set_options_invalid(option, value):
    old = OPTIONS[option]
    with pytest.raises(ValueError):
        set_options(**{option: value})
    assert old == OPTIONS[option]


def test_register_missing_method():
    @register_missing_method("test")
    class MissingTest(MissingBase):
        def is_missing(self, null, count, a_param=2):
            return True

        @staticmethod
        def validate(a_param):
            return a_param < 3

    with pytest.raises(ValueError):
        set_options(missing_options={"test": {"a_param": 5}})

    with set_options(check_missing="test"):
        assert OPTIONS["check_missing"] == "test"
